//
// libsyd: Rust-based C library for syd interaction via /dev/syd
// lib/src/lib.rs: syd API C Library
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: LGPL-3.0-or-later

//! # libsyd - syd API Rust Library
//!
//! `libsyd` is a C library written in Rust that implements the syd
//! stat API, providing an interface to the `/dev/syd` of syd. It
//! allows for runtime configuration and interaction with the syd
//! sandboxing environment.
//!
//! ## Overview
//! The library is designed to interact with the syd sandboxing
//! environment, offering functionalities to check and modify the state
//! of the sandbox lock, and perform system calls to `/dev/syd`.
//!
//! For more detailed information and usage instructions, refer to the syd
//! manual, available at [syd Manual](http://man.exherbolinux.org/syd.2.html).
//!
//! ## Author
//! Ali Polatel <alip@chesswob.org>

// We like safe, clean and simple code with documentation.
#![deny(missing_docs)]
#![deny(clippy::allow_attributes_without_reason)]
#![deny(clippy::arithmetic_side_effects)]
#![deny(clippy::as_ptr_cast_mut)]
#![deny(clippy::as_underscore)]
#![deny(clippy::assertions_on_result_states)]
#![deny(clippy::borrow_as_ptr)]
#![deny(clippy::branches_sharing_code)]
#![deny(clippy::case_sensitive_file_extension_comparisons)]
#![deny(clippy::cast_lossless)]
#![deny(clippy::cast_possible_truncation)]
#![deny(clippy::cast_possible_wrap)]
#![deny(clippy::cast_precision_loss)]
#![deny(clippy::cast_ptr_alignment)]
#![deny(clippy::cast_sign_loss)]
#![deny(clippy::checked_conversions)]
#![deny(clippy::clear_with_drain)]
#![deny(clippy::clone_on_ref_ptr)]
#![deny(clippy::cloned_instead_of_copied)]
#![deny(clippy::cognitive_complexity)]
#![deny(clippy::collection_is_never_read)]
#![deny(clippy::copy_iterator)]
#![deny(clippy::create_dir)]
#![deny(clippy::dbg_macro)]
#![deny(clippy::debug_assert_with_mut_call)]
#![deny(clippy::decimal_literal_representation)]
#![deny(clippy::default_trait_access)]
#![deny(clippy::default_union_representation)]
#![deny(clippy::derive_partial_eq_without_eq)]
#![deny(clippy::doc_link_with_quotes)]
#![deny(clippy::doc_markdown)]
#![deny(clippy::explicit_into_iter_loop)]
#![deny(clippy::explicit_iter_loop)]
#![deny(clippy::fallible_impl_from)]
#![deny(clippy::missing_safety_doc)]
#![deny(clippy::undocumented_unsafe_blocks)]

use std::{
    ffi::{CStr, OsStr, OsString},
    fs::{symlink_metadata, Metadata},
    os::{
        raw::{c_char, c_int},
        unix::{
            ffi::OsStrExt,
            fs::{FileTypeExt, MetadataExt},
        },
    },
    path::{Path, PathBuf},
};

/// An enumeration of the possible states for the sandbox lock.
#[repr(u8)]
#[allow(non_camel_case_types)]
pub enum lock_state_t {
    /// The sandbox lock is off, allowing all sandbox commands.
    LOCK_OFF,
    /// The sandbox lock is set to on for all processes except the initial
    /// process (syd exec child). This is the default state.
    LOCK_EXEC,
    /// The sandbox lock is on, disallowing all sandbox commands.
    LOCK_ON,
}

/// An enumeration of the possible actions for sandboxing.
#[repr(u8)]
#[allow(non_camel_case_types)]
pub enum action_t {
    /// Allow system call.
    ALLOW,
    /// Allow system call and warn.
    WARN,
    /// Deny system call silently.
    FILTER,
    /// Deny system call and warn.
    DENY,
    /// Deny system call, warn and stop the offending process.
    STOP,
    /// Deny system call, warn and kill the offending process.
    KILL,
    /// Warn, and exit Syd immediately with deny errno as exit value.
    EXIT,
}

const EFAULT: i32 = 14;
const EINVAL: i32 = 22;

#[inline(always)]
fn check_stat(stat: &Metadata) -> bool {
    if !stat.file_type().is_char_device() {
        return false;
    }

    let rdev = stat.rdev();

    let major = (rdev >> 8) & 0xff;
    let minor = rdev & 0xff;

    // dev/null
    major == 1 && minor == 3
}

fn stat<P: AsRef<Path>>(path: P) -> c_int {
    match symlink_metadata(path) {
        Ok(stat) if check_stat(&stat) => 0,
        Ok(_) => -EINVAL,
        Err(error) => match error.raw_os_error() {
            Some(e) => e.checked_neg().unwrap_or(-EINVAL),
            None => -EINVAL,
        },
    }
}

fn esyd<P: AsRef<Path>>(rule: P, elem: *const c_char, op: u8) -> c_int {
    if !matches!(op, b'+' | b'-' | b'^' | b':') {
        return -EINVAL;
    }

    if elem.is_null() {
        return -EFAULT;
    }

    // SAFETY: Trust that `elem` is a null-terminated string.
    let elem = unsafe { CStr::from_ptr(elem) };
    let elem = OsStr::from_bytes(elem.to_bytes());

    // Manually concatenate the path segments
    let mut path = OsString::from("/dev/syd/");
    path.push(rule.as_ref());
    path.push(OsStr::from_bytes(&[op]));
    path.push(elem);

    // Convert the OsString to PathBuf
    let path = PathBuf::from(path);

    stat(path)
}

/// Performs a syd API check
///
/// The caller is advised to perform this check before
/// calling any other syd API calls.
///
/// Returns API number on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_api() -> c_int {
    match stat("/dev/syd/3") {
        0 => 3,
        n => n,
    }
}

/// Performs an lstat system call on the file "/dev/syd".
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_check() -> c_int {
    stat("/dev/syd")
}

/// Causes syd to exit immediately with code 127
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_panic() -> c_int {
    stat("/dev/syd/panic")
}

/// Causes syd to reset sandboxing to the default state.
/// Allowlists, denylists and filters are going to be cleared.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_reset() -> c_int {
    stat("/dev/syd/reset")
}

/// Causes syd to read configuration from the given file descriptor.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_load(fd: c_int) -> c_int {
    let mut path = PathBuf::from("/dev/syd/load");

    let mut buf = itoa::Buffer::new();
    path.push(buf.format(fd));

    stat(path)
}

/// Sets the state of the sandbox lock.
///
/// state: The desired state of the sandbox lock.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_lock(state: lock_state_t) -> c_int {
    match state as u8 {
        0 => stat("/dev/syd/lock:off"),
        1 => stat("/dev/syd/lock:exec"),
        2 => stat("/dev/syd/lock:on"),
        _ => -EINVAL,
    }
}

/// Checks if exec sandboxing is enabled.
///
/// Returns true if exec sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_exec() -> bool {
    stat("/dev/syd/sandbox/exec?") == 0
}

/// Enable exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_exec() -> c_int {
    stat("/dev/syd/sandbox/exec:on")
}

/// Disable exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_exec() -> c_int {
    stat("/dev/syd/sandbox/exec:off")
}

/// Checks if read sandboxing is enabled.
///
/// Returns true if read sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_read() -> bool {
    stat("/dev/syd/sandbox/read?") == 0
}

/// Enable read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_read() -> c_int {
    stat("/dev/syd/sandbox/read:on")
}

/// Disable read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_read() -> c_int {
    stat("/dev/syd/sandbox/read:off")
}

/// Checks if stat sandboxing is enabled.
///
/// Returns true if stat sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_stat() -> bool {
    stat("/dev/syd/sandbox/stat?") == 0
}

/// Enable stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_stat() -> c_int {
    stat("/dev/syd/sandbox/stat:on")
}

/// Disable stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_stat() -> c_int {
    stat("/dev/syd/sandbox/stat:off")
}

/// Checks if write sandboxing is enabled.
///
/// Returns true if write sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_write() -> bool {
    stat("/dev/syd/sandbox/write?") == 0
}

/// Enable write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_write() -> c_int {
    stat("/dev/syd/sandbox/write:on")
}

/// Disable write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_write() -> c_int {
    stat("/dev/syd/sandbox/write:off")
}

/// Checks if ioctl sandboxing is enabled.
///
/// Returns true if ioctl sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_ioctl() -> bool {
    stat("/dev/syd/sandbox/ioctl?") == 0
}

/// Enable ioctl sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_ioctl() -> c_int {
    stat("/dev/syd/sandbox/ioctl:on")
}

/// Disable ioctl sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_ioctl() -> c_int {
    stat("/dev/syd/sandbox/ioctl:off")
}

/// Checks if net sandboxing is enabled.
///
/// Returns true if net sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_net() -> bool {
    stat("/dev/syd/sandbox/net?") == 0
}

/// Enable net sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_net() -> c_int {
    stat("/dev/syd/sandbox/net:on")
}

/// Disable net sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_net() -> c_int {
    stat("/dev/syd/sandbox/net:off")
}

/// Checks if force sandboxing is enabled.
///
/// Returns true if force sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_force() -> bool {
    stat("/dev/syd/sandbox/force?") == 0
}

/// Enable force sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_force() -> c_int {
    stat("/dev/syd/sandbox/force:on")
}

/// Disable force sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_force() -> c_int {
    stat("/dev/syd/sandbox/force:off")
}

/// Checks if TPE sandboxing is enabled.
///
/// Returns true if TPE sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_tpe() -> bool {
    stat("/dev/syd/sandbox/tpe?") == 0
}

/// Enable TPE sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_tpe() -> c_int {
    stat("/dev/syd/sandbox/tpe:on")
}

/// Disable TPE sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_tpe() -> c_int {
    stat("/dev/syd/sandbox/tpe:off")
}

/// Checks if memory sandboxing is enabled.
///
/// Returns true if memory sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_mem() -> bool {
    stat("/dev/syd/sandbox/mem?") == 0
}

/// Enable memory sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_mem() -> c_int {
    stat("/dev/syd/sandbox/mem:on")
}

/// Disable memory sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_mem() -> c_int {
    stat("/dev/syd/sandbox/mem:off")
}

/// Checks if PID sandboxing is enabled.
///
/// Returns true if PID sandboxing is enabled, false otherwise.
#[no_mangle]
pub extern "C" fn syd_enabled_pid() -> bool {
    stat("/dev/syd/sandbox/pid?") == 0
}

/// Enable PID sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_enable_pid() -> c_int {
    stat("/dev/syd/sandbox/pid:on")
}

/// Disable PID sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_disable_pid() -> c_int {
    stat("/dev/syd/sandbox/pid:off")
}

/// Set the default action for Exec Sandboxing.
#[no_mangle]
pub extern "C" fn syd_default_exec(action: action_t) -> c_int {
    // Convert action_t enum to corresponding action string.
    stat(match action {
        action_t::ALLOW => "/dev/syd/default/exec:allow",
        action_t::WARN => "/dev/syd/default/exec:warn",
        action_t::FILTER => "/dev/syd/default/exec:filter",
        action_t::DENY => "/dev/syd/default/exec:deny",
        action_t::STOP => "/dev/syd/default/exec:stop",
        action_t::KILL => "/dev/syd/default/exec:kill",
        action_t::EXIT => "/dev/syd/default/exec:exit",
    })
}

/// Set the default action for Read Sandboxing.
#[no_mangle]
pub extern "C" fn syd_default_read(action: action_t) -> c_int {
    // Convert action_t enum to corresponding action string.
    stat(match action {
        action_t::ALLOW => "/dev/syd/default/read:allow",
        action_t::WARN => "/dev/syd/default/read:warn",
        action_t::FILTER => "/dev/syd/default/read:filter",
        action_t::DENY => "/dev/syd/default/read:deny",
        action_t::STOP => "/dev/syd/default/read:stop",
        action_t::KILL => "/dev/syd/default/read:kill",
        action_t::EXIT => "/dev/syd/default/read:exit",
    })
}

/// Set the default action for Write Sandboxing.
#[no_mangle]
pub extern "C" fn syd_default_write(action: action_t) -> c_int {
    // Convert action_t enum to corresponding action string.
    stat(match action {
        action_t::ALLOW => "/dev/syd/default/write:allow",
        action_t::WARN => "/dev/syd/default/write:warn",
        action_t::FILTER => "/dev/syd/default/write:filter",
        action_t::DENY => "/dev/syd/default/write:deny",
        action_t::STOP => "/dev/syd/default/write:stop",
        action_t::KILL => "/dev/syd/default/write:kill",
        action_t::EXIT => "/dev/syd/default/write:exit",
    })
}

/// Set the default action for Stat Sandboxing.
#[no_mangle]
pub extern "C" fn syd_default_stat(action: action_t) -> c_int {
    // Convert action_t enum to corresponding action string.
    stat(match action {
        action_t::ALLOW => "/dev/syd/default/stat:allow",
        action_t::WARN => "/dev/syd/default/stat:warn",
        action_t::FILTER => "/dev/syd/default/stat:filter",
        action_t::DENY => "/dev/syd/default/stat:deny",
        action_t::STOP => "/dev/syd/default/stat:stop",
        action_t::KILL => "/dev/syd/default/stat:kill",
        action_t::EXIT => "/dev/syd/default/stat:exit",
    })
}

/// Set the default action for Ioctl Sandboxing.
#[no_mangle]
pub extern "C" fn syd_default_ioctl(action: action_t) -> c_int {
    // Convert action_t enum to corresponding action string.
    stat(match action {
        action_t::ALLOW => "/dev/syd/default/ioctl:allow",
        action_t::WARN => "/dev/syd/default/ioctl:warn",
        action_t::FILTER => "/dev/syd/default/ioctl:filter",
        action_t::DENY => "/dev/syd/default/ioctl:deny",
        action_t::STOP => "/dev/syd/default/ioctl:stop",
        action_t::KILL => "/dev/syd/default/ioctl:kill",
        action_t::EXIT => "/dev/syd/default/ioctl:exit",
    })
}

/// Set the default action for Network Sandboxing.
#[no_mangle]
pub extern "C" fn syd_default_net(action: action_t) -> c_int {
    // Convert action_t enum to corresponding action string.
    stat(match action {
        action_t::ALLOW => "/dev/syd/default/net:allow",
        action_t::WARN => "/dev/syd/default/net:warn",
        action_t::FILTER => "/dev/syd/default/net:filter",
        action_t::DENY => "/dev/syd/default/net:deny",
        action_t::STOP => "/dev/syd/default/net:stop",
        action_t::KILL => "/dev/syd/default/net:kill",
        action_t::EXIT => "/dev/syd/default/net:exit",
    })
}

/// Set the default action for Memory Sandboxing.
#[no_mangle]
pub extern "C" fn syd_default_mem(action: action_t) -> c_int {
    // Convert action_t enum to corresponding action string.
    stat(match action {
        action_t::ALLOW => "/dev/syd/default/mem:allow",
        action_t::WARN => "/dev/syd/default/mem:warn",
        action_t::FILTER => "/dev/syd/default/mem:filter",
        action_t::DENY => "/dev/syd/default/mem:deny",
        action_t::STOP => "/dev/syd/default/mem:stop",
        action_t::KILL => "/dev/syd/default/mem:kill",
        action_t::EXIT => "/dev/syd/default/mem:exit",
    })
}

/// Set the default action for PID Sandboxing.
#[no_mangle]
pub extern "C" fn syd_default_pid(action: action_t) -> c_int {
    // Convert action_t enum to corresponding action string.
    stat(match action {
        action_t::ALLOW => "/dev/syd/default/pid:allow",
        action_t::WARN => "/dev/syd/default/pid:warn",
        action_t::FILTER => "/dev/syd/default/pid:filter",
        action_t::DENY => "/dev/syd/default/pid:deny",
        action_t::STOP => "/dev/syd/default/pid:stop",
        action_t::KILL => "/dev/syd/default/pid:kill",
        action_t::EXIT => "/dev/syd/default/pid:exit",
    })
}

/// Set the default action for TPE Sandboxing.
#[no_mangle]
pub extern "C" fn syd_default_tpe(action: action_t) -> c_int {
    // Convert action_t enum to corresponding action string.
    stat(match action {
        action_t::ALLOW => "/dev/syd/default/tpe:allow",
        action_t::WARN => "/dev/syd/default/tpe:warn",
        action_t::FILTER => "/dev/syd/default/tpe:filter",
        action_t::DENY => "/dev/syd/default/tpe:deny",
        action_t::STOP => "/dev/syd/default/tpe:stop",
        action_t::KILL => "/dev/syd/default/tpe:kill",
        action_t::EXIT => "/dev/syd/default/tpe:exit",
    })
}

/// Set the default action for Force Sandboxing.
#[no_mangle]
pub extern "C" fn syd_default_force(action: action_t) -> c_int {
    // Convert action_t enum to corresponding action string.
    stat(match action {
        action_t::ALLOW => "/dev/syd/default/force:allow",
        action_t::WARN => "/dev/syd/default/force:warn",
        action_t::FILTER => "/dev/syd/default/force:filter",
        action_t::DENY => "/dev/syd/default/force:deny",
        action_t::STOP => "/dev/syd/default/force:stop",
        action_t::KILL => "/dev/syd/default/force:kill",
        action_t::EXIT => "/dev/syd/default/force:exit",
    })
}

/// Set the default action for SegvGuard
#[no_mangle]
pub extern "C" fn syd_default_segvguard(action: action_t) -> c_int {
    // Convert action_t enum to corresponding action string.
    stat(match action {
        action_t::ALLOW => "/dev/syd/default/segvguard:allow",
        action_t::WARN => "/dev/syd/default/segvguard:warn",
        action_t::FILTER => "/dev/syd/default/segvguard:filter",
        action_t::DENY => "/dev/syd/default/segvguard:deny",
        action_t::STOP => "/dev/syd/default/segvguard:stop",
        action_t::KILL => "/dev/syd/default/segvguard:kill",
        action_t::EXIT => "/dev/syd/default/segvguard:exit",
    })
}

/// Adds a request to the _ioctl_(2) denylist.
#[no_mangle]
pub extern "C" fn syd_ioctl_deny(request: u64) -> c_int {
    stat(&format!("/dev/syd/ioctl/deny+{request}"))
}

/// Adds an entry to the Integrity Force map for Force Sandboxing.
///
/// # Safety
///
/// This function is marked `unsafe` because it dereferences raw
/// pointers, which is inherently unsafe in Rust.
///
/// The caller must ensure the following conditions are met to safely
/// use this function:
///
/// 1. The `path` pointer must point to a valid, null-terminated C-style
///    string.
/// 2. The `hash` pointer must point to a valid, null-terminated C-style
///    string.
#[no_mangle]
pub unsafe extern "C" fn syd_force_add(
    path: *const c_char,
    hash: *const c_char,
    action: action_t,
) -> c_int {
    if path.is_null() || hash.is_null() {
        return -EFAULT;
    }

    // SAFETY: Trust that `path` and `hash` are a null-terminated strings.
    let path = unsafe { CStr::from_ptr(path) };
    // SAFETY: ditto
    let hash = unsafe { CStr::from_ptr(hash) };
    let path = match path.to_str() {
        Ok(s) => s,
        Err(_) => return -EINVAL,
    };
    let hash = match hash.to_str() {
        Ok(s) => s,
        Err(_) => return -EINVAL,
    };

    // Convert action_t enum to corresponding action string.
    let action = match action {
        action_t::WARN => "warn",
        action_t::STOP => "stop",
        action_t::KILL => "kill",
        _ => return -EINVAL,
    };

    // Call the stat function with the formatted string.
    stat(format!("/dev/syd/force+{path}:{hash}:{action}"))
}

/// Removes an entry from the Integrity Force map for Force Sandboxing.
/// # Safety
///
/// This function is marked `unsafe` because it dereferences raw
/// pointers, which is inherently unsafe in Rust.
///
/// The caller must ensure the following conditions are met to safely
/// use this function:
///
/// 1. The `path` pointer must point to a valid, null-terminated C-style
///    string.
#[no_mangle]
pub unsafe extern "C" fn syd_force_del(path: *const c_char) -> c_int {
    if path.is_null() {
        return -EFAULT;
    }

    // SAFETY: Trust that `path` is a null-terminated string.
    let path = unsafe { CStr::from_ptr(path) };
    let path = match path.to_str() {
        Ok(s) => s,
        Err(_) => return -EINVAL,
    };

    // Call the stat function with the formatted string.
    stat(format!("/dev/syd/force-{path}"))
}

/// Clears the Integrity Force map for Force Sandboxing.
#[no_mangle]
pub extern "C" fn syd_force_clr() -> c_int {
    stat("/dev/syd/force^")
}

/// Adds to the given actionlist of exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_exec_add(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/exec", glob, b'+'),
        action_t::WARN => esyd("warn/exec", glob, b'+'),
        action_t::FILTER => esyd("filter/exec", glob, b'+'),
        action_t::DENY => esyd("deny/exec", glob, b'+'),
        action_t::STOP => esyd("stop/exec", glob, b'+'),
        action_t::KILL => esyd("kill/exec", glob, b'+'),
        action_t::EXIT => esyd("exit/exec", glob, b'+'),
    }
}

/// Removes the first instance from the end of the given actionlist of
/// exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_exec_del(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/exec", glob, b'-'),
        action_t::WARN => esyd("warn/exec", glob, b'-'),
        action_t::FILTER => esyd("filter/exec", glob, b'-'),
        action_t::DENY => esyd("deny/exec", glob, b'-'),
        action_t::STOP => esyd("stop/exec", glob, b'-'),
        action_t::KILL => esyd("kill/exec", glob, b'-'),
        action_t::EXIT => esyd("exit/exec", glob, b'-'),
    }
}

/// Removes all matching patterns from the given actionlist of exec sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_exec_rem(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/exec", glob, b'^'),
        action_t::WARN => esyd("warn/exec", glob, b'^'),
        action_t::FILTER => esyd("filter/exec", glob, b'^'),
        action_t::DENY => esyd("deny/exec", glob, b'^'),
        action_t::STOP => esyd("stop/exec", glob, b'^'),
        action_t::KILL => esyd("kill/exec", glob, b'^'),
        action_t::EXIT => esyd("exit/exec", glob, b'^'),
    }
}

/// Adds to the given actionlist of read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_read_add(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/read", glob, b'+'),
        action_t::WARN => esyd("warn/read", glob, b'+'),
        action_t::FILTER => esyd("filter/read", glob, b'+'),
        action_t::DENY => esyd("deny/read", glob, b'+'),
        action_t::STOP => esyd("stop/read", glob, b'+'),
        action_t::KILL => esyd("kill/read", glob, b'+'),
        action_t::EXIT => esyd("exit/read", glob, b'+'),
    }
}

/// Removes the first instance from the end of the given actionlist of
/// read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_read_del(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/read", glob, b'-'),
        action_t::WARN => esyd("warn/read", glob, b'-'),
        action_t::FILTER => esyd("filter/read", glob, b'-'),
        action_t::DENY => esyd("deny/read", glob, b'-'),
        action_t::STOP => esyd("stop/read", glob, b'-'),
        action_t::KILL => esyd("kill/read", glob, b'-'),
        action_t::EXIT => esyd("exit/read", glob, b'-'),
    }
}

/// Removes all matching patterns from the given actionlist of read sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_read_rem(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/read", glob, b'^'),
        action_t::WARN => esyd("warn/read", glob, b'^'),
        action_t::FILTER => esyd("filter/read", glob, b'^'),
        action_t::DENY => esyd("deny/read", glob, b'^'),
        action_t::STOP => esyd("stop/read", glob, b'^'),
        action_t::KILL => esyd("kill/read", glob, b'^'),
        action_t::EXIT => esyd("exit/read", glob, b'^'),
    }
}

/// Adds to the given actionlist of write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_write_add(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/write", glob, b'+'),
        action_t::WARN => esyd("warn/write", glob, b'+'),
        action_t::FILTER => esyd("filter/write", glob, b'+'),
        action_t::DENY => esyd("deny/write", glob, b'+'),
        action_t::STOP => esyd("stop/write", glob, b'+'),
        action_t::KILL => esyd("kill/write", glob, b'+'),
        action_t::EXIT => esyd("exit/write", glob, b'+'),
    }
}

/// Removes the first instance from the end of the given actionlist of
/// write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_write_del(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/write", glob, b'-'),
        action_t::WARN => esyd("warn/write", glob, b'-'),
        action_t::FILTER => esyd("filter/write", glob, b'-'),
        action_t::DENY => esyd("deny/write", glob, b'-'),
        action_t::STOP => esyd("stop/write", glob, b'-'),
        action_t::KILL => esyd("kill/write", glob, b'-'),
        action_t::EXIT => esyd("exit/write", glob, b'-'),
    }
}

/// Removes all matching patterns from the given actionlist of write sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_write_rem(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/write", glob, b'^'),
        action_t::WARN => esyd("warn/write", glob, b'^'),
        action_t::FILTER => esyd("filter/write", glob, b'^'),
        action_t::DENY => esyd("deny/write", glob, b'^'),
        action_t::STOP => esyd("stop/write", glob, b'^'),
        action_t::KILL => esyd("kill/write", glob, b'^'),
        action_t::EXIT => esyd("exit/write", glob, b'^'),
    }
}

/// Adds to the given actionlist of stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_stat_add(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/stat", glob, b'+'),
        action_t::WARN => esyd("warn/stat", glob, b'+'),
        action_t::FILTER => esyd("filter/stat", glob, b'+'),
        action_t::DENY => esyd("deny/stat", glob, b'+'),
        action_t::STOP => esyd("stop/stat", glob, b'+'),
        action_t::KILL => esyd("kill/stat", glob, b'+'),
        action_t::EXIT => esyd("exit/stat", glob, b'+'),
    }
}

/// Removes the first instance from the end of the given actionlist of
/// stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_stat_del(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/stat", glob, b'-'),
        action_t::WARN => esyd("warn/stat", glob, b'-'),
        action_t::FILTER => esyd("filter/stat", glob, b'-'),
        action_t::DENY => esyd("deny/stat", glob, b'-'),
        action_t::STOP => esyd("stop/stat", glob, b'-'),
        action_t::KILL => esyd("kill/stat", glob, b'-'),
        action_t::EXIT => esyd("exit/stat", glob, b'-'),
    }
}

/// Removes all matching patterns from the given actionlist of stat sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_stat_rem(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/stat", glob, b'^'),
        action_t::WARN => esyd("warn/stat", glob, b'^'),
        action_t::FILTER => esyd("filter/stat", glob, b'^'),
        action_t::DENY => esyd("deny/stat", glob, b'^'),
        action_t::STOP => esyd("stop/stat", glob, b'^'),
        action_t::KILL => esyd("kill/stat", glob, b'^'),
        action_t::EXIT => esyd("exit/stat", glob, b'^'),
    }
}

/// Adds to the given actionlist of ioctl sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_ioctl_add(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/ioctl", glob, b'+'),
        action_t::WARN => esyd("warn/ioctl", glob, b'+'),
        action_t::FILTER => esyd("filter/ioctl", glob, b'+'),
        action_t::DENY => esyd("deny/ioctl", glob, b'+'),
        action_t::STOP => esyd("stop/ioctl", glob, b'+'),
        action_t::KILL => esyd("kill/ioctl", glob, b'+'),
        action_t::EXIT => esyd("exit/ioctl", glob, b'+'),
    }
}

/// Removes the first instance from the end of the given actionlist of
/// ioctl sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_ioctl_del(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/ioctl", glob, b'-'),
        action_t::WARN => esyd("warn/ioctl", glob, b'-'),
        action_t::FILTER => esyd("filter/ioctl", glob, b'-'),
        action_t::DENY => esyd("deny/ioctl", glob, b'-'),
        action_t::STOP => esyd("stop/ioctl", glob, b'-'),
        action_t::KILL => esyd("kill/ioctl", glob, b'-'),
        action_t::EXIT => esyd("exit/ioctl", glob, b'-'),
    }
}

/// Removes all matching patterns from the given actionlist of ioctl sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_ioctl_rem(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/ioctl", glob, b'^'),
        action_t::WARN => esyd("warn/ioctl", glob, b'^'),
        action_t::FILTER => esyd("filter/ioctl", glob, b'^'),
        action_t::DENY => esyd("deny/ioctl", glob, b'^'),
        action_t::STOP => esyd("stop/ioctl", glob, b'^'),
        action_t::KILL => esyd("kill/ioctl", glob, b'^'),
        action_t::EXIT => esyd("exit/ioctl", glob, b'^'),
    }
}

/// Adds to the given actionlist of net/bind sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_net_bind_add(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/net/bind", glob, b'+'),
        action_t::WARN => esyd("warn/net/bind", glob, b'+'),
        action_t::FILTER => esyd("filter/net/bind", glob, b'+'),
        action_t::DENY => esyd("deny/net/bind", glob, b'+'),
        action_t::STOP => esyd("stop/net/bind", glob, b'+'),
        action_t::KILL => esyd("kill/net/bind", glob, b'+'),
        action_t::EXIT => esyd("exit/net/bind", glob, b'+'),
    }
}

/// Removes the first instance from the end of the given actionlist of
/// net/bind sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_net_bind_del(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/net/bind", glob, b'-'),
        action_t::WARN => esyd("warn/net/bind", glob, b'-'),
        action_t::FILTER => esyd("filter/net/bind", glob, b'-'),
        action_t::DENY => esyd("deny/net/bind", glob, b'-'),
        action_t::STOP => esyd("stop/net/bind", glob, b'-'),
        action_t::KILL => esyd("kill/net/bind", glob, b'-'),
        action_t::EXIT => esyd("exit/net/bind", glob, b'-'),
    }
}

/// Removes all matching patterns from the given actionlist of net/bind sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_net_bind_rem(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/net/bind", glob, b'^'),
        action_t::WARN => esyd("warn/net/bind", glob, b'^'),
        action_t::FILTER => esyd("filter/net/bind", glob, b'^'),
        action_t::DENY => esyd("deny/net/bind", glob, b'^'),
        action_t::STOP => esyd("stop/net/bind", glob, b'^'),
        action_t::KILL => esyd("kill/net/bind", glob, b'^'),
        action_t::EXIT => esyd("exit/net/bind", glob, b'^'),
    }
}

/// Adds to the given actionlist of net/connect sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_net_connect_add(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/net/connect", glob, b'+'),
        action_t::WARN => esyd("warn/net/connect", glob, b'+'),
        action_t::FILTER => esyd("filter/net/connect", glob, b'+'),
        action_t::DENY => esyd("deny/net/connect", glob, b'+'),
        action_t::STOP => esyd("stop/net/connect", glob, b'+'),
        action_t::KILL => esyd("kill/net/connect", glob, b'+'),
        action_t::EXIT => esyd("exit/net/connect", glob, b'+'),
    }
}

/// Removes the first instance from the end of the given actionlist of
/// net/connect sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_net_connect_del(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/net/connect", glob, b'-'),
        action_t::WARN => esyd("warn/net/connect", glob, b'-'),
        action_t::FILTER => esyd("filter/net/connect", glob, b'-'),
        action_t::DENY => esyd("deny/net/connect", glob, b'-'),
        action_t::STOP => esyd("stop/net/connect", glob, b'-'),
        action_t::KILL => esyd("kill/net/connect", glob, b'-'),
        action_t::EXIT => esyd("exit/net/connect", glob, b'-'),
    }
}

/// Removes all matching patterns from the given actionlist of net/connect sandboxing.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_net_connect_rem(action: action_t, glob: *const c_char) -> c_int {
    match action {
        action_t::ALLOW => esyd("allow/net/connect", glob, b'^'),
        action_t::WARN => esyd("warn/net/connect", glob, b'^'),
        action_t::FILTER => esyd("filter/net/connect", glob, b'^'),
        action_t::DENY => esyd("deny/net/connect", glob, b'^'),
        action_t::STOP => esyd("stop/net/connect", glob, b'^'),
        action_t::KILL => esyd("kill/net/connect", glob, b'^'),
        action_t::EXIT => esyd("exit/net/connect", glob, b'^'),
    }
}

/// Set syd maximum per-process memory usage limit for memory sandboxing.
///
/// parse-size crate is used to parse the value so formatted strings are OK.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_mem_max(size: *const c_char) -> c_int {
    esyd("mem/max", size, b':')
}

/// Set syd maximum per-process virtual memory usage limit for memory sandboxing.
///
/// parse-size crate is used to parse the value so formatted strings are OK.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_mem_vm_max(size: *const c_char) -> c_int {
    esyd("mem/vm_max", size, b':')
}

/// Set syd maximum process id limit for PID sandboxing
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_pid_max(size: usize) -> c_int {
    let mut path = OsString::from("/dev/syd/pid/max:");

    let mut buf = itoa::Buffer::new();
    let max_str = OsStr::from_bytes(buf.format(size).as_bytes());
    path.push(max_str);

    stat(path)
}

/// Specify SegvGuard entry expiry timeout in seconds.
/// Setting this timeout to 0 effectively disables SegvGuard.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_segvguard_expiry(timeout: u64) -> c_int {
    let mut path = OsString::from("/dev/syd/segvguard/expiry:");

    let mut buf = itoa::Buffer::new();
    let exp_str = OsStr::from_bytes(buf.format(timeout).as_bytes());
    path.push(exp_str);

    stat(path)
}

/// Specify SegvGuard entry suspension timeout in seconds.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_segvguard_suspension(timeout: u64) -> c_int {
    let mut path = OsString::from("/dev/syd/segvguard/suspension:");

    let mut buf = itoa::Buffer::new();
    let exp_str = OsStr::from_bytes(buf.format(timeout).as_bytes());
    path.push(exp_str);

    stat(path)
}

/// Specify SegvGuard max number of crashes before suspension.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub extern "C" fn syd_segvguard_maxcrashes(max: u8) -> c_int {
    let mut path = OsString::from("/dev/syd/segvguard/maxcrashes:");

    let mut buf = itoa::Buffer::new();
    let max_str = OsStr::from_bytes(buf.format(max).as_bytes());
    path.push(max_str);

    stat(path)
}

/// Execute a command outside the sandbox without sandboxing
///
/// # Safety
///
/// This function is marked `unsafe` because it dereferences raw
/// pointers, which is inherently unsafe in Rust.
///
/// The caller must ensure the following conditions are met to safely
/// use this function:
///
/// 1. The `file` pointer must point to a valid, null-terminated C-style
///    string.
///
/// 2. The `argv` pointer must point to an array of pointers, where each
///    pointer refers to a valid, null-terminated C-style string. The
///    last pointer in the array must be null, indicating the end of the
///    array.
///
/// 3. The memory pointed to by `file` and `argv` must remain valid for
///    the duration of the call.
///
/// Failing to uphold these guarantees can lead to undefined behavior,
/// including memory corruption and data races.
///
/// Returns 0 on success, negated errno on failure.
#[no_mangle]
pub unsafe extern "C" fn syd_exec(file: *const c_char, argv: *const *const c_char) -> c_int {
    if file.is_null() || argv.is_null() {
        return -EFAULT;
    }

    // SAFETY: Trust that `file` is a null-terminated string.
    let file = CStr::from_ptr(file);
    let file = OsStr::from_bytes(file.to_bytes());

    let mut path = OsString::from("/dev/syd/cmd/exec!");
    path.push(file);

    let mut idx: isize = 0;
    while !(*argv.offset(idx)).is_null() {
        // SAFETY: Trust that each `argv` element is a null-terminated string.
        let arg = CStr::from_ptr(*argv.offset(idx));
        let arg = OsStr::from_bytes(arg.to_bytes());

        path.push(OsStr::from_bytes(&[b'\x1F'])); // ASCII Unit Separator
        path.push(arg);

        idx = idx.saturating_add(1);
    }

    let path = PathBuf::from(path);
    stat(path)
}
