;;; -*- lexical-binding: t -*-
;;;
;;; syd.el --- Emacs Lisp implementation of the syd stat interface
;;;
;;; Syd: rock-solid unikernel
;;; lib/src/syd.el: Emacs Lisp implementation of the syd stat interface
;;;
;;; Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
;;;
;;; SPDX-License-Identifier: GPL-3.0-or-later

                                        ; Define lock states as keywords
(defconst syd-lock-off :lock-off
  "The sandbox lock is off, allowing all sandbox commands.")

(defconst syd-lock-exec :lock-exec
  "The sandbox lock is set to on for all processes except the initial process
\(syd exec child). This is the default state.")

(defconst syd-lock-on :lock-on
  "The sandbox lock is on, disallowing all sandbox commands.")

                                        ; Define sandbox actions as keywords
(defconst syd-action-allow :action-allow
  "Allow system call.")

(defconst syd-action-warn :action-warn
  "Allow system call and warn.")

(defconst syd-action-filter :action-filter
  "Deny system call silently.")

(defconst syd-action-deny :action-deny
  "Deny system call and warn.")

(defconst syd-action-stop :action-stop
  "Deny system call, warn and stop offending process.")

(defconst syd-action-kill :action-kill
  "Deny system call, warn and kill offending process.")

(defconst syd-action-exit :action-exit
  "Warn, and exit Syd immediately with deny errno as exit value.")

(defun syd-info ()
  "Reads the state of the syd sandbox from /dev/syd and returns it as an alist.
If the `json' module is not available, returns nil."
  (if (require 'json nil t)
      (condition-case nil
          (with-temp-buffer
            (insert-file-contents "/dev/syd")
            (with-no-warnings
              (let ((json-object-type 'alist)
                    (json-array-type 'list)
                    (json-key-type 'symbol)
                    (json-false nil)
                    (json-null nil))
                (json-read))))
        (file-error
         (message "Error reading /dev/syd.")
         nil)
        (json-error
         (message "JSON decoding error.")
         nil))
    (progn
      (message "JSON module not available.")
      nil)))

(defun syd-api ()
  "Performs a syd API check."
  (if (syd--stat "/dev/syd/3")
      3   ; API number on success
    nil)) ; On error, return nil

(defun syd-check ()
  "Check if '/dev/syd' is a character device."
  (syd--stat "/dev/syd"))

(defun syd-panic ()
  "Causes syd to exit immediately with code 127"
  (syd--stat "/dev/syd/panic"))

(defun syd-reset ()
  "Causes syd to reset sandboxing to the default state."
  (syd--stat "/dev/syd/reset"))

(defun syd-load (fd)
  "Causes syd to read configuration from the given file descriptor FD."
  (let ((path (concat "/dev/syd/load/" (number-to-string fd))))
    (syd--stat path)))

(defun syd-lock (state)
  "Sets the state of the sandbox lock.
STATE is one of the keywords :lock-off, :lock-exec, or :lock-on.
Returns t on success, nil on failure."
  (cond
   ((eq state syd-lock-off) (syd--stat "/dev/syd/lock:off"))
   ((eq state syd-lock-exec) (syd--stat "/dev/syd/lock:exec"))
   ((eq state syd-lock-on) (syd--stat "/dev/syd/lock:on"))
   (t nil))) ; Invalid state

(defun syd-enabled-mem ()
  "Checks if memory sandboxing is enabled."
  (syd--stat "/dev/syd/sandbox/mem?"))

(defun syd-enable-mem ()
  "Enable memory sandboxing."
  (syd--stat "/dev/syd/sandbox/mem:on"))

(defun syd-disable-mem ()
  "Disable memory sandboxing."
  (syd--stat "/dev/syd/sandbox/mem:off"))

(defun syd-enabled-pid ()
  "Checks if PID sandboxing is enabled."
  (syd--stat "/dev/syd/sandbox/pid?"))

(defun syd-enable-pid ()
  "Enable PID sandboxing."
  (syd--stat "/dev/syd/sandbox/pid:on"))

(defun syd-disable-pid ()
  "Disable PID sandboxing."
  (syd--stat "/dev/syd/sandbox/pid:off"))

(defun syd-enabled-tpe ()
  "Checks if TPE sandboxing is enabled."
  (syd--stat "/dev/syd/sandbox/tpe?"))

(defun syd-enable-tpe ()
  "Enable TPE sandboxing."
  (syd--stat "/dev/syd/sandbox/tpe:on"))

(defun syd-disable-tpe ()
  "Disable TPE sandboxing."
  (syd--stat "/dev/syd/sandbox/tpe:off"))

(defun syd-enabled-read ()
  "Checks if Read sandboxing is enabled."
  (syd--stat "/dev/syd/sandbox/read?"))

(defun syd-enable-read ()
  "Enable Read sandboxing."
  (syd--stat "/dev/syd/sandbox/read:on"))

(defun syd-disable-read ()
  "Disable Read sandboxing."
  (syd--stat "/dev/syd/sandbox/read:off"))

(defun syd-enabled-stat ()
  "Checks if Stat sandboxing is enabled."
  (syd--stat "/dev/syd/sandbox/stat?"))

(defun syd-enable-stat ()
  "Enable Stat sandboxing."
  (syd--stat "/dev/syd/sandbox/stat:on"))

(defun syd-disable-stat ()
  "Disable Stat sandboxing."
  (syd--stat "/dev/syd/sandbox/stat:off"))

(defun syd-enabled-write ()
  "Checks if Write sandboxing is enabled."
  (syd--stat "/dev/syd/sandbox/write?"))

(defun syd-enable-write ()
  "Enable Write sandboxing."
  (syd--stat "/dev/syd/sandbox/write:on"))

(defun syd-disable-write ()
  "Disable Write sandboxing."
  (syd--stat "/dev/syd/sandbox/write:off"))

(defun syd-enabled-ioctl ()
  "Checks if Ioctl sandboxing is enabled."
  (syd--stat "/dev/syd/sandbox/ioctl?"))

(defun syd-enable-ioctl ()
  "Enable Ioctl sandboxing."
  (syd--stat "/dev/syd/sandbox/ioctl:on"))

(defun syd-disable-ioctl ()
  "Disable Ioctl sandboxing."
  (syd--stat "/dev/syd/sandbox/ioctl:off"))

(defun syd-enabled-exec ()
  "Checks if Exec sandboxing is enabled."
  (syd--stat "/dev/syd/sandbox/exec?"))

(defun syd-enable-exec ()
  "Enable Exec sandboxing."
  (syd--stat "/dev/syd/sandbox/exec:on"))

(defun syd-disable-exec ()
  "Disable Exec sandboxing."
  (syd--stat "/dev/syd/sandbox/exec:off"))

(defun syd-enabled-force ()
  "Checks if force sandboxing is enabled."
  (syd--stat "/dev/syd/sandbox/force?"))

(defun syd-enable-force ()
  "Enable force sandboxing."
  (syd--stat "/dev/syd/sandbox/force:on"))

(defun syd-disable-force ()
  "Disable force sandboxing."
  (syd--stat "/dev/syd/sandbox/force:off"))

(defun syd-enabled-net ()
  "Checks if Network sandboxing is enabled."
  (syd--stat "/dev/syd/sandbox/net?"))

(defun syd-enable-net ()
  "Enable Network sandboxing."
  (syd--stat "/dev/syd/sandbox/net:on"))

(defun syd-disable-net ()
  "Disable Network sandboxing."
  (syd--stat "/dev/syd/sandbox/net:off"))

(defun syd-default-read (action)
  "Set default action for Read sandboxing.
ACTION is a constant representing the sandboxing action."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-kill) "kill")
                 ((eq action :action-exit) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      (let ((cmd (format "/dev/syd/default/read:%s" action)))
        ;; Call syd--stat with the command
        (syd--stat cmd)))))

(defun syd-default-stat (action)
  "Set default action for Stat sandboxing.
ACTION is a constant representing the sandboxing action."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-kill) "kill")
                 ((eq action :action-exit) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      (let ((cmd (format "/dev/syd/default/stat:%s" action)))
        ;; Call syd--stat with the command
        (syd--stat cmd)))))

(defun syd-default-write (action)
  "Set default action for Write sandboxing.
ACTION is a constant representing the sandboxing action."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-kill) "kill")
                 ((eq action :action-exit) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      (let ((cmd (format "/dev/syd/default/write:%s" action)))
        ;; Call syd--write with the command
        (syd--stat cmd)))))

(defun syd-default-ioctl (action)
  "Set default action for Ioctl sandboxing.
ACTION is a constant representing the sandboxing action."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-kill) "kill")
                 ((eq action :action-exit) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      (let ((cmd (format "/dev/syd/default/ioctl:%s" action)))
        ;; Call syd--ioctl with the command
        (syd--stat cmd)))))

(defun syd-default-exec (action)
  "Set default action for Exec sandboxing.
ACTION is a constant representing the sandboxing action."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-kill) "kill")
                 ((eq action :action-exit) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      (let ((cmd (format "/dev/syd/default/exec:%s" action)))
        ;; Call syd--exec with the command
        (syd--stat cmd)))))

(defun syd-default-net (action)
  "Set default action for Network sandboxing.
ACTION is a constant representing the sandboxing action."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-kill) "kill")
                 ((eq action :action-exit) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      (let ((cmd (format "/dev/syd/default/net:%s" action)))
        ;; Call syd--net with the command
        (syd--stat cmd)))))

(defun syd-default-mem (action)
  "Set default action for Memory sandboxing.
ACTION is a constant representing the sandboxing action."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-kill) "kill")
                 ((eq action :action-exit) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      (let ((cmd (format "/dev/syd/default/mem:%s" action)))
        ;; Call syd--net with the command
        (syd--stat cmd)))))

(defun syd-default-pid (action)
  "Set default action for PID sandboxing.
ACTION is a constant representing the sandboxing action."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-kill) "kill")
                 ((eq action :action-exit) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      (let ((cmd (format "/dev/syd/default/pid:%s" action)))
        ;; Call syd--net with the command
        (syd--stat cmd)))))

(defun syd-default-tpe (action)
  "Set default action for TPE sandboxing.
ACTION is a constant representing the sandboxing action."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-kill) "kill")
                 ((eq action :action-exit) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      (let ((cmd (format "/dev/syd/default/tpe:%s" action)))
        ;; Call syd--net with the command
        (syd--stat cmd)))))

(defun syd-default-force (action)
  "Set default action for Force sandboxing.
ACTION is a constant representing the sandboxing action."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-kill) "kill")
                 ((eq action :action-exit) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      (let ((cmd (format "/dev/syd/default/force:%s" action)))
        ;; Call syd--net with the command
        (syd--stat cmd)))))

(defun syd-default-segvguard (action)
  "Set default action for SegvGuard.
ACTION is a constant representing the sandboxing action."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-kill) "kill")
                 ((eq action :action-exit) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      (let ((cmd (format "/dev/syd/default/segvguard:%s" action)))
        ;; Call syd--net with the command
        (syd--stat cmd)))))

(defun syd-ioctl-deny (request)
  "Adds a request to the _ioctl_(2) denylist.
REQUEST is the _ioctl_(2) request number to add to the denylist."
  (unless (numberp request)
    (error "Request must be a number"))
  (let ((path (format "/dev/syd/ioctl/deny+%d" request)))
    (syd--stat path)))

(defun syd-force-add (path hash action)
  "Adds an entry to the Integrity Force map for Force Sandboxing.
PATH is a fully-qualified file name.
HASH is a hexadecimal encoded checksum.
ACTION is one of :action-warn, :action-stop or :action-kill."
  (let ((action (cond ((eq action :action-warn) "warn")
                      ((eq action :action-stop) "stop")
                      ((eq action :action-kill) "kill"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "/dev/syd/force+%s:%s:%s" path hash action)))
        ;; Call syd--stat with the command
        (syd--stat cmd)))))

(defun syd-force-del (path)
  "Removes an entry from the Integrity Force map for Force Sandboxing.
PATH is a fully-qualified file name."
  ;; Create the command string
  (let ((cmd (format "/dev/syd/force-%s" path)))
    ;; Call syd--stat with the command
    (syd--stat cmd)))

(defun syd-force-clr ()
  "Clears the Integrity Force map for Force Sandboxing."
  (syd--stat "/dev/syd/force^"))

(defun syd-read-add (action glob)
  "Adds to the given actionlist of read sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/read" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?+))))))

(defun syd-read-del (action glob)
  "Removes the first matching entry from the end of the given actionlist of read sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/read" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?-))))))

(defun syd-read-rem (action glob)
  "Removes all matching entries from the given actionlist of read sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/read" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?^))))))

(defun syd-stat-add (action glob)
  "Adds to the given actionlist of stat sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/stat" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?+))))))

(defun syd-stat-del (action glob)
  "Removes the first matching entry from the end of the given actionlist of stat sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/stat" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?-))))))

(defun syd-stat-rem (action glob)
  "Removes all matching entries from the given actionlist of stat sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/stat" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?^))))))

(defun syd-write-add (action glob)
  "Adds to the given actionlist of write sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/write" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?+))))))

(defun syd-write-del (action glob)
  "Removes the first matching entry from the end of the given actionlist of write sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/write" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?-))))))

(defun syd-write-rem (action glob)
  "Removes all matching entries from the given actionlist of write sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/write" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?^))))))

(defun syd-ioctl-add (action glob)
  "Adds to the given actionlist of ioctl sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/ioctl" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?+))))))

(defun syd-ioctl-del (action glob)
  "Removes the first matching entry from the end of the given actionlist of ioctl sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/ioctl" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?-))))))

(defun syd-ioctl-rem (action glob)
  "Removes all matching entries from the given actionlist of ioctl sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/ioctl" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?^))))))

(defun syd-exec-add (action glob)
  "Adds to the given actionlist of exec sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/exec" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?+))))))

(defun syd-exec-del (action glob)
  "Removes the first matching entry from the end of the given actionlist of exec sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/exec" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?-))))))

(defun syd-exec-rem (action glob)
  "Removes all matching entries from the given actionlist of exec sandboxing.
ACTION is a constant representing the sandboxing action.
GLOB is a string representing the glob pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/exec" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd glob ?^))))))

(defun syd-net-bind-add (action addr)
  "Adds to the given actionlist of exec sandboxing.
ACTION is a constant representing the sandboxing action.
ADDR is a string representing the address pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/net/bind" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd addr ?+))))))

(defun syd-net-bind-del (action addr)
  "Removes the first matching entry from the end of the given actionlist of net/bind sandboxing.
ACTION is a constant representing the sandboxing action.
ADDR is a string representing the address pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/net/bind" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd addr ?-))))))

(defun syd-net-bind-rem (action addr)
  "Removes all matching entries from the given actionlist of net/bind sandboxing.
ACTION is a constant representing the sandboxing action.
ADDR is a string representing the address pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/net/bind" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd addr ?^))))))

(defun syd-net-connect-add (action addr)
  "Adds to the given actionlist of exec sandboxing.
ACTION is a constant representing the sandboxing action.
ADDR is a string representing the address pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/net/connect" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd addr ?+))))))

(defun syd-net-connect-del (action addr)
  "Removes the first matching entry from the end of the given actionlist of net/connect sandboxing.
ACTION is a constant representing the sandboxing action.
ADDR is a string representing the address pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/net/connect" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd addr ?-))))))

(defun syd-net-connect-rem (action addr)
  "Removes all matching entries from the given actionlist of net/connect sandboxing.
ACTION is a constant representing the sandboxing action.
ADDR is a string representing the address pattern."
  (let ((action (cond 
                 ((eq action :action-allow) "allow")
                 ((eq action :action-warn) "warn")
                 ((eq action :action-filter) "filter")
                 ((eq action :action-deny) "deny")
                 ((eq action :action-stop) "stop")
                 ((eq action :action-stop) "kill")
                 ((eq action :action-kill) "exit"))))
    ;; Only proceed if action is not nil
    (when action
      ;; Create the command string
      (let ((cmd (format "%s/net/connect" action)))
        ;; Call syd--stat with the command
        (syd--stat (syd--rule cmd addr ?^))))))

(defun syd-mem-max (size)
  "Set syd maximum per-process memory usage limit.
SIZE can be an integer or a string representing the memory limit."
  (let ((size-str (cond ((integerp size) (number-to-string size))
                        ((stringp size) size)
                        (t (error "Size must be an integer or a string")))))
    (syd--stat (syd--rule "mem/max" size-str ?:))))

(defun syd-mem-vm-max (size)
  "Set syd maximum per-process virtual memory usage limit.
SIZE can be an integer or a string representing the memory limit."
  (let ((size-str (cond ((integerp size) (number-to-string size))
                        ((stringp size) size)
                        (t (error "Size must be an integer or a string")))))
    (syd--stat (syd--rule "mem/vm_max" size-str ?:))))

(defun syd-pid-max (size)
  "Set syd maximum process ID limit for PID sandboxing.
SIZE is a number representing the PID limit."
  (unless (numberp size)
    (error "Size must be a number"))
  (let ((path (format "/dev/syd/pid/max:%d" size)))
    (syd--stat path)))

(defun syd-segvguard-expiry (timeout)
  "Specify SegvGuard entry expiry timeout in seconds.
Setting this timeout to 0 effectively disables SegvGuard.
TIMEOUT is a number representing the timeout in seconds."
  (unless (numberp timeout)
    (error "Timeout must be a number"))
  (let ((path (format "/dev/syd/segvguard/expiry:%d" timeout)))
    (syd--stat path)))

(defun syd-segvguard-suspension (timeout)
  "Specify SegvGuard entry suspension timeout in seconds.
TIMEOUT is a number representing the timeout in seconds."
  (unless (numberp timeout)
    (error "Timeout must be a number"))
  (let ((path (format "/dev/syd/segvguard/suspension:%d" timeout)))
    (syd--stat path)))

(defun syd-segvguard-maxcrashes (limit)
  "Specify SegvGuard max number of crashes before suspension.
LIMIT is a number representing the crash limit."
  (unless (numberp limit)
    (error "Limit must be a number"))
  (let ((path (format "/dev/syd/segvguard/maxcrashes:%d" limit)))
    (syd--stat path)))

(defun syd-exec (file argv)
  "Execute a command outside the sandbox without sandboxing.
FILE is the file path of the command as a string.
ARGV is a list of strings representing the arguments to the command."
  (unless (stringp file)
    (error "File must be a string"))
  (let ((all-strings t))
    (dolist (arg argv)
      (unless (stringp arg)
        (setq all-strings nil)))
    (unless all-strings
      (error "All elements in ARGV must be strings")))

  (let ((cmd (mapconcat 'identity (cons file argv) "\x1F")))
    (syd--stat (concat "/dev/syd/cmd/exec!" cmd))))

(defun syd--rule (rule elem op)
  "Helper function to construct a path for syd operations.
RULE is a string representing the rule.
ELEM is a string representing the element.
OP is a character representing the operation."
  (unless (member op '(?+ ?- ?^ ?:))
    (error "Invalid operation"))
  (when (string-empty-p elem)
    (error "Element cannot be empty"))
  (concat "/dev/syd/" rule (char-to-string op) elem))

(defun syd--stat (path)
  "Check if the file at PATH exists using `file-modes'."
  (condition-case nil
      (not (null (file-modes path)))
    (error nil)))  ; On error, return nil

(provide 'syd)
;;; syd.el ends here
