SYD(2)

# NAME

*/dev/syd* virtual system call interface

# SYNOPSIS

```
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* Execute sandbox commands */
int stat("/dev/syd/[command]", NULL);         // "syd -m command"
int stat("/dev/syd/[config]?", NULL);
int stat("/dev/syd/[config]:[value]", NULL);  // "syd -m config:value"
int stat("/dev/syd/[list]+[value]", NULL);    // "syd -m list+value"
int stat("/dev/syd/[list]-[value]", NULL);    // "syd -m list-value"
int stat("/dev/syd/[list]^[value]", NULL);    // "syd -m list^value"
int stat("/dev/syd/[command]![value]", NULL);

/* Read sandbox state as JSON */
int open("/dev/syd", O_RDONLY);
```

# DESCRIPTION

The */dev/syd* virtual system call interface is a unique mechanism designed for
runtime configuration of the Syd sandbox environment. It enables sandboxed
processes to interact with the Syd process to dynamically adjust sandbox
settings or query its state. This interaction is facilitated through the use of
virtual system calls, specifically via the _stat_(2) system call, applied to
specially constructed paths under */dev/syd*.

This interface allows for a range of operations, including enabling or disabling
sandbox features, appending or removing elements from lists, querying the
sandbox state, and executing special Syd commands. Operations are specified
through paths constructed with */dev/syd* as the prefix, followed by a sandbox
command and an operation character that denotes the desired action:

- *:* for setting a value (boolean, integer, string),
- *?* for querying a value,
- *+* for appending to a string array,
- *-* for removing an element from a string array,
- *^* for removing all matching elements from a string array, and
- *!* for executing a special Syd command.

This interface supports a flexible and powerful method for managing sandbox
policies dynamically, allowing for real-time adjustments to the security and
operational behavior of sandboxed processes. *libsyd* is a comprehensive C
library designed for interfacing with the Syd stat interface. It offers
functionalities for managing sandbox states, and facilitating runtime
configuration and interaction with the Syd sandboxing environment. *gosyd* is
a Go module that uses _cgo_ to use *libsyd*. *plsyd* is a Perl module that uses
_FFI::Platypus_ to use *libsyd*. *pysyd* is a Python module that uses _ctypes_
to use *libsyd*. *rbsyd* is a Ruby module that uses _ffi_ gem to use *libsyd*.
*syd.el* is an _Emacs Lisp_ implementation of the Syd stat interface.

# COMMANDS

The */dev/syd* interface supports the following commands for runtime
configuration of the sandbox. Each command can be invoked through the _stat_(2)
system call on special paths under */dev/syd*. Note, Syd provides similar
interfaces for the _stat_(2) interface, *-m* command-line option and the
configuration file. Some sandbox commands only take affect when they're
submitted on startup, such as *unshare/user*, and *sandbox/lock*. Such commands
are noted as *static* in the descriptions below.

## stat

This command causes Syd to output sandbox state on standard error.

## reset

This command causes Syd to reset sandboxing to the default state.
Allowlists, denylists and filters are going to be cleared. The state of
the sandbox lock is not affected by reset. This ensures an unintended
reset cannot open window for a sandbox bypass.

## panic

This command causes Syd to exit immediately with code 127.

Due to security reasons, this command is only available via the virtual stat
call, it may not be used with the *-m* command line switch or in a configuration
file.

## ghost

This command initiates Ghost mode. Ghost mode is irreversible so you can
call this command only once during Syd runtime. See Ghost mode section
in _syd_(7) manual page for more information. In case of successful initiation,
the virtual _stat_(2) call returns -1 and sets the errno to "EOWNERDEAD".

Due to security reasons, this command implies "reset", ie. the sandbox
state is reset before Ghost mode initiation to ensure there're no
run-away exec processes after the invocation of the "ghost" command.

Due to security reasons, this command is only available via the virtual stat
call, it may not be used with the *-m* command line switch or in a configuration
file.

## lock

[[ type
:[ *string*

Set the state of the sandbox lock. Possible values are *on*, *off* and *exec*.
If the sandbox lock is *on* no sandbox commands are allowed. If *exec* is
specified, the sandbox lock is set to *on* for all processes except the initial
process, aka Syd exec child. Note, the sandbox lock used to default to *exec*
but as a hardening measure and to ensure security by default, as of version
3.17.0, this has been changed such that the default is _unset_ and if no lock
clause has been specified by the time Syd executes the initial sandbox process,
then the sandbox lock is automatically set to *on*. This means if no "lock"
clause is specified in any of the profiles, configuration files or "-m" CLI
arguments, the lock will be *on* by default. Setting lock to *on* at any point
during configuration parsing prevents further commands from being emitted by the
sandbox. This feature may be used to lock site-wide defaults for a Syd login
shell by adding a "lock:on" clause at the end of the site-wide configuration
file which prevents Syd from subsequently parsing the user configuration file,
practically enforcing the site-wide defaults.

Note, setting lock to *off* or *exec* at startup makes Syd skip preventing
_execve_(2) and _execveat_(2) system calls as part of the "Execution Control
(EEC)" feature. This is done to allow "cmd/exec" command to execute commands
outside the sandbox.

## sandbox/exec

[[ type
:[ *boolean*
|  default
:  *on*
|  query
:  *yes*

Turn Exec sandboxing *on* or *off*.

## sandbox/read

[[ type
:[ *boolean*
|  default
:  *on*
|  query
:  *yes*

Turn Read sandboxing *on* or *off*.

## sandbox/write

[[ type
:[ *boolean*
|  default
:  *on*
|  query
:  *yes*

Turn Write sandboxing *on* or *off*.

## sandbox/stat

[[ type
:[ *boolean*
|  default
:  *on*
|  query
:  *yes*

Turn Stat sandboxing *on* or *off*.

## sandbox/ioctl

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Turn Ioctl sandboxing *on* or *off*.

For performance reasons, this only works at startup. If not given at
startup, Syd will just allow _ioctl_(2) system calls at seccomp-bpf
level, and all _ioctl_(2) requests _except those in the denylist_ will be
allowed regardless of path. Turning this sandboxing off during runtime
is still possible, in this case the _ioctl_(2) system call handler will
only check the _ioctl_(2) request against the allowlist and the denylist
but will not check the path to the _ioctl_(2) file descriptor for
access.

## sandbox/net

[[ type
:[ *boolean*
|  default
:  *on*
|  query
:  *yes*

Turn Network sandboxing *on* or *off*.

## sandbox/lock

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Turn Landlock sandboxing *on* or *off*.

## sandbox/mem

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Turn Memory sandboxing *on* or *off*.

For performance reasons, this only works at startup. If not given at startup,
Syd will just allow _brk_(2), _mmap_(2), _mmap2_(2), and _mremap_(2) system
calls at seccomp-bpf level. Turning this sandboxing off during runtime is still
possible, in this case the respective system calls handlers will do nothing and
just continue the calls.

## sandbox/pid

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Turn PID sandboxing *on* or *off*.

For performance reasons, this only works at startup. If not given at startup,
Syd will just allow _fork_(2), _vfork_(2), _clone_(2), and _clone3_(2) system
calls at seccomp-bpf level. Turning this sandboxing off during runtime is still
possible, in this case the respective system calls will do nothing and just
continue the calls.

## sandbox/proxy

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Turn Proxy sandboxing *on* or *off*.

Defaults to proxying through TOR. See the options "proxy/addr",
"proxy/port", "proxy/ext/addr", and "proxy/ext/port" to configure a
different proxy.

Implies *unshare/net:1*.

## sandbox/force

[[ type
:[ *boolean*
|  query
:  *yes*

Turn Force sandboxing *on* or *off*.

## sandbox/tpe

[[ type
:[ *boolean*
|  default
:  *on*
|  query
:  *yes*

Turn Trusted Path Execution (TPE) sandboxing *on* or *off*.

## default/exec

[[ type
:[ *string*
|  default
:  *deny*

Specify the default action for Exec sandboxing access violations.
The value must be exactly one of "allow", "warn", "filter", "deny", "stop", "kill", or "exit".

## default/read

[[ type
:[ *string*
|  default
:  *deny*

Specify the default action for Read sandboxing access violations.
The value must be exactly one of "allow", "warn", "filter", "deny", "stop", "kill", or "exit".

## default/write

[[ type
:[ *string*
|  default
:  *deny*

Specify the default action for Write sandboxing access violations.
The value must be exactly one of "allow", "warn", "filter", "deny", "stop", "kill", or "exit".

## default/stat

[[ type
:[ *string*
|  default
:  *deny*

Specify the default action for Stat sandboxing access violations.
The value must be exactly one of "allow", "warn", "filter", "deny", "stop", "kill", or "exit".

## default/ioctl

[[ type
:[ *string*
|  default
:  *deny*

Specify the default action for Ioctl sandboxing access violations.
The value must be exactly one of "allow", "warn", "filter", "deny", "stop", "kill", or "exit".

## default/net

[[ type
:[ *string*
|  default
:  *deny*

Specify the default action for Network sandboxing access violations.
The value must be exactly one of "allow", "warn", "filter", "deny", "stop", "kill", or "exit".

## default/mem

[[ type
:[ *string*
|  default
:  *deny*

Specify the action for Memory sandboxing access violations.
The value must be exactly one of "warn", "filter", "deny", "stop", "kill", or "exit".

## default/pid

[[ type
:[ *string*
|  default
:  *kill*

Specify the action for PID sandboxing access violations.
The value must be exactly one of "warn", "filter", "deny", "stop", "kill", or "exit".

## default/force

[[ type
:[ *string*
|  default
:  *kill*

For force sandboxing, define the default action to take when the path of a
binary is not in the Integrity Force map. The value must be either one of
*allow*, *warn*, *filter*, *deny*, *stop* or *kill* where the default is *kill*.

## default/tpe

[[ type
:[ *string*
|  default
:  *deny*

Specify the action for TPE sandboxing access violations.
The value must be exactly one of "warn", "filter", "deny", "stop", "kill", or "exit".

## default/segvguard

[[ type
:[ *string*
|  default
:  *kill*

Specify the action for SegvGuard access violations.
The value must be exactly one of "warn", "filter", "deny", "stop", "kill", or "exit".
The values "deny" and "kill" are functionally equivalent.

## unshare/mount

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Create Mount namespace on startup.

## unshare/uts

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Create UTS namespace on startup.

## unshare/ipc

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Create IPC namespace on startup.

## unshare/user

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Create User namespace on startup.

## unshare/pid

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Create Pid namespace on startup.

## unshare/net

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Create Net namespace on startup.

## unshare/cgroup

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Create CGroup namespace on startup.

## unshare/time

[[ type
:[ *boolean*
|  query
:  *yes*
|  static
:  *yes*

Create Time namespace on startup. Syd resets the boot-time clock
such that _uptime_(1) will report container uptime rather than
host uptime.

## root

[[ type
:[ *string*
|  static
:  *yes*

Change directory and chroot to root before starting the process. The
directory *root/proc* _must_ exist to mount private proc filesystem.

Implies *unshare/mount:1* and *unshare/pid:1*.

As of version 3.23.14, symbolic links are not followed in any part of
the root directory and path traversal using ".." is not permitted. In
addition, root directory must be an absolute path, relative paths are
not permitted.

## root/map

[[ type
:[ *boolean*
|  static
:  *yes*

Map current user to root in the sandbox on startup.

Implies *unshare/user:1*.

## root/fake

[[ type
:[ *boolean*
|  static
:  *yes*

In *fakeroot* mode, the system will return a user/group id of *0*,
mimicking the *root* user. This allows users to execute commands with
apparent root privileges, without actual superuser rights. It's useful
for tasks like package building where root-like environment is needed,
but not actual root permissions.

## name/host

[[ type
:[ *string*
|  default
:  *<auto>*
|  static
:  *yes*

Set host name in the sandbox. Implies *unshare/uts:1*.

## name/domain

[[ type
:[ *string*
|  default
:  *syd-3*
|  static
:  *yes*

Set NIS/YP domain name in the sandbox. Implies *unshare/uts:1*.

## time

[[ type
:[ *time_t*
|  static
:  *yes*

Set clock monotonic and boottime offset (seconds) in Time Namespace.
Implies *unshare/time:1*.

## ioctl/allow

[[ type
:[ *positive integer-array (u64)*
|  default
:  *[...]*

Add to or remove a request from the _ioctl_(2) request allowlist.
Accepts an unsigned 64-bit integer as argument. Prefix with "0x" for
hexadecimal and "0o" for octal input. Use "ioctl/allow+<request>" to add
to, and "ioctl/allow-<request>" to remove from the allowlist. By
default the list contains the following _ioctl_(2) requests:

. FIBMAP
. FICLONE
. FICLONERANGE
. FIDEDUPERANGE
. FIOASYNC
. FIOCLEX
. FIOGETOWN
. FIONBIO
. FIONCLEX
. FIONREAD
. FIOQSIZE
. FIOSETOWN
. FS_IOC_FIEMAP
. FS_IOC_GETFLAGS
. GIO_UNIMAP
. SIOCGPGRP
. SIOCSPGRP
. TCFLSH
. TCGETA
. TCGETS
. TCSBRK
. TCSBRKP
. TCSETA
. TCSETAF
. TCSETAW
. TCSETS
. TCSETSF
. TCSETSW
. TCXONC
. TIOCCBRK
. TIOCGETD
. TIOCGPGRP
. TIOCINQ
. TIOCOUTQ
. TIOCSBRK

Note, for rules added at startup deny rules have precedence over allow
rules because the denylist is checked at kernel-space, whereas the
allowlist is checked at user-space. For rules added after startup, the
last matching rule wins.

## ioctl/deny

[[ type
:[ *positive integer-array (u64)*
|  default
:  *[...]*
|  static
:  *add is dynamic, remove is partly static*

Add to or remove a request from the _ioctl_(2) request denylist. Accepts
an unsigned 64-bit integer as argument. Prefix with "0x" for hexadecimal
and "0o" for octal input. Use "ioctl/deny+<request>" to add to, and
"ioctl/deny-<request>" to remove from the allowlist. By default the
list of denylisted _ioctl_(2) requests are as follows:

. FS_IOC_FSGETXATTR
. FS_IOC_FSSETXATTR
. FS_IOC_SETFLAGS
. TIOCCONS
. TIOCLINUX
. TIOCSETD
. TIOCSTI

Note, for security reasons, the _ioctl_(2) denylist is applied at the
parent seccomp-bpf filter at startup. This means the Syd process is
included in this restriction as well. This also means, removing elements
from this list after startup has no effect. However, if Ioctl sandboxing
was enabled at startup, adding new elements to the _ioctl_(2) denylist
will further restrict the _ioctl_(2) request space.

Note, for rules added at startup deny rules have precedence over allow
rules because the denylist is checked at kernel-space, whereas the
allowlist is checked at user-space. For rules added after startup, the
last matching rule wins.

Further reading about denylisted _ioctl_(2) requests:

. https://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-2023-1523
. https://a13xp0p0v.github.io/2017/03/24/CVE-2017-2636.html
. http://phrack.org/issues/52/6.html#article

## mem/max

[[ type
:[ *positive integer (usize)*
|  default
:  *128MB*

This setting specifies the limit on per-process memory usage. Setting this value
to *0* disables testing for this type of memory usage. Note, the value is parsed
using the *parse-size* crate. Refer to their documentation for information on
formatting.

## mem/vm_max

[[ type
:[ *positive integer (usize)*
|  default
:  *0*

This setting specifies the limit on per-process virtual memory usage. Setting
this value to *0* disables testing for this type of memory usage. Note, the
value is parsed using the *parse-size* crate. Refer to their documentation for
information on formatting.

## pid/max

[[ type
:[ *positive integer (usize)*
|  default
:  *128*

This setting specifies the limit on the number of running tasks for pid
sandboxing. Setting this value to *0* is functionally equivalent to setting
*sandbox/pid* to *off*.

## bind

[[ type
:[ *string-array*
|  static
:  *yes*

This command causes Syd to bind mount a directory on startup. The
format is *source-dir:target-dir:mount-options,...* where _the source and
target directories may be equal_. Mount options are a comma-separated list
of a combination of the following options:

- "ro" to mount the filesystem read-only.
- "nodev" to not interpret character or block special devices on the filesystem.
- "noexec" to not permit direct execution of any binaries on the mounted \
filesystem.
- "nosuid" to not honour set-user-ID and set-group-ID bits or file capabilities \
when executing programs from this filesystem. In addition, SELinux domain \
transitions require permission "nosuid_transition", which in turn needs also \
policy capability "nnp_nosuid_transition".
- "noatime" to not update inode access times on this filesystem (e.g. for \
faster access on the news spool to speed up news servers). This works for all \
inode types (directories too), so it implies "nodiratime".
- "nodiratime" to not update directory inode access times on this filesystem. \
(This option is implied when "noatime" is set.)
- "relatime" to update inode access times relative to modify or change time.

Note, this command implies "unshare/mount:1". Mount options may be omitted.
If the source directory does not start with a slash, it is interpreted as the
filesystem type rather than the source directory. This may be used to mount
special filesystems such as cgroupfs, overlayfs or _tmpfs_(5) into the mount
namespace. In this case, any mount options supported by this filesystem type may
be submitted in options argument not just the ones listed above. You may find
some examples below:

- bind+/:/:ro
- bind+tmpfs:/tmp:noexec,size=16M
- bind+cgroup2:/sys/fs/cgroup:nodev,noexec,nosuid
- bind+overlay:/tmp/target:lowerdir=/tmp/lower,upperdir=/tmp/upper,workdir=/tmp/work,nosuid

This command may be used to create immutable containers. For example, the
command "bind+/:/:ro" is functionally equivalent to "deny/write+/\*\*\*" except
the restriction happens at _kernel VFS layer_ rather than at user level using
_seccomp-notify_. Alternatively this can also be achieved at the kernel level
using _Landlock_.

Note, if the source or target directory does not exist, Syd logs an
"info" level entry and continues processing the rest of the mounts. Any
other error during bind mounting is fatal.

As of version 3.23.14, symbolic links are not followed in any part of
the source or target directory paths and path traversal using ".." is
not permitted. In addition, target directory must be an absolute path,
relative paths are not permitted.

As of version 3.23.14, mounting the special "proc" filesystem under a
custom path is not permitted. Syd handles this mount itself specially
after all bind mounts are processed.

## proxy/addr

[[ type
:[ *IP address*
|  default
:  *127.0.0.1*
|  static
:  *yes*

Set internal address for Proxy sandboxing. This must be an IPv4 or an
IPv6 address. Defaults to 127.0.0.1.

## proxy/port

[[ type
:[ *integer*
|  default
:  *9050*
|  static
:  *yes*

Set internal port for Proxy sandboxing. Defaults to 9050.

## proxy/ext/addr

[[ type
:[ *IP address*
|  default
:  *127.0.0.1*
|  static
:  *yes*

Set external address for Proxy sandboxing. This must be an IPv4 or an
IPv6 address. Defaults to 127.0.0.1.

## proxy/ext/port

[[ type
:[ *integer*
|  default
:  *9050*
|  static
:  *yes*

Set external port for Proxy sandboxing. Defaults to 9050.

## force

[[ type
:[ *string-array*

Add or remove an integrity force rule for Force Sandboxing. The format is
_force+/path:hashhex:action_ for addition and _force-/path_ for removal. Use
_force^_ to clear the Integrity Force map. Available actions are the following:

- kill: Send the offending process SIGKILL signal.
- stop: Send the offending process SIGSTOP signal.
- warn: Allow execution.

In all cases a warning to the syslog is written. hashhex is either a
32-character MD5 checksum, a 40-character SHA1 checksum, a 64-character SHA3-256
checksum, a 96-character SHA3-384 checksum or a 128-character SHA3-512 checksum.

- _syd-sha_(1) is a helper tool to calculate checksums of files.
- _syd-path_(1) is a helper tool to write integrity force rules for binaries under PATH.

## segvguard/expiry

[[ type
:[ *integer (u64)*
|  default
:  120

Specify SegvGuard expiry timeout in seconds. Set to 0 to disable SegvGuard.

## segvguard/suspension

[[ type
:[ *integer (u64)*
|  default
:  600

Specify SegvGuard suspension timeout in seconds.

## segvguard/maxcrashes

[[ type
:[ *integer (u8)*
|  default
:  5

Specify SegvGuard max crashes.

## tpe/gid

[[ type
:[ *integer (uid_t)*

Specify untrusted GID for Trusted Path Execution (TPE). By default, TPE
is applied to users of all groups including root and this setting can be
used to limit it to a certain group. To unset a previously set GID and
return to the default state set "none" as the value.

## tpe/negate

[[ type
:[ *boolean*

Negate GID logic for Trusted Path Execution (TPE). This turns "tpe/gid"
from untrusted into trusted such that users belonging to this group will
be exempt from TPE.

## tpe/root_owned

[[ type
:[ *boolean*

Ensure directory is root-owned for Trusted Path Execution (TPE).

Note, this option will misbehave with "unshare/user:1" if the real root
user is not mapped inside the container.

## tpe/user_owned

[[ type
:[ *boolean*

Ensure directory is user-owned or root for Trusted Path Execution (TPE).

Note, this option will misbehave with "unshare/user:1" if the real root
is not mapped inside the container.

## allow/lock/read

[[ type
:[ *string-array*
|  static
:  *yes*
|  default
:  _["/proc"]_

Specifies a list of beneath paths to allow as read-only for *Landlock
sandboxing*.

As of version 3.21.0, this list includes the path "/proc" by default as
Syd is included in the Landlock sandbox and Syd requires access to this
directory to function correctly.

## allow/lock/write

[[ type
:[ *string-array*
|  static
:  *yes*
|  default
:  _["/dev/null"]_

Specifies a list of beneath paths to allow as read-write for *Landlock
sandboxing*.

As of version 3.21.0, this list includes the path "/dev/null" by default
as Syd is included in the Landlock sandbox and Syd requires access to
this file to function correctly.

## allow/exec

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to allow for *Exec sandboxing*.

## allow/read

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to allow for *Read sandboxing*.

## allow/stat

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to allow for *Stat sandboxing*.

## allow/write

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to allow for *Write sandboxing*.

## allow/ioctl

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to allow for *Ioctl sandboxing*.

## allow/net/bind

[[ type
:[ *string-array*

Specifies a list of network address patterns to allow for *Bind network sandboxing*.

## allow/net/connect

[[ type
:[ *string-array*

Specifies a list of network address patterns to allow for *Connect network sandboxing*.

## allow/net/link

[[ type
:[ *string-array*
|  static
:  *yes*

Specifies a list of netlink families to allow for *Network sandboxing*.
Accepts a comma-delimited list of the following items: "route",
"usersock", "firewall", "sock_diag", "nflog", "xfrm", "selinux",
"iscsi", "audit", "fib_lookup", "connector", "netfilter", "ip6_fw",
"dnrtmsg", "kobject_uevent", "generic", "scsitransport", "ecryptfs",
"rdma" and "crypto".

## warn/exec

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to warn for *Exec sandboxing*.

## warn/read

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to warn for *Read sandboxing*.

## warn/write

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to warn for *Write sandboxing*.

## warn/stat

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to warn for *Stat sandboxing*.

## warn/ioctl

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to warn for *Ioctl sandboxing*.

## warn/net/bind

[[ type
:[ *string-array*

Specifies a list of network address patterns to warn for *Bind network sandboxing*.

## warn/net/connect

[[ type
:[ *string-array*

Specifies a list of network address patterns to warn for *Connect network sandboxing*.

## deny/exec

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to deny for *Exec sandboxing*.

## deny/read

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to deny for *Read sandboxing*.

## deny/write

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to deny for *Write sandboxing*.

## deny/stat

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to deny for *Stat sandboxing*.

## deny/ioctl

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to deny for *Ioctl sandboxing*.

## deny/net/bind

[[ type
:[ *string-array*

Specifies a list of network address patterns to deny for *Bind network sandboxing*.

## deny/net/connect

[[ type
:[ *string-array*

Specifies a list of network address patterns to deny for *Connect network sandboxing*.

## filter/exec

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to filter for *Exec sandboxing*.

## filter/read

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to filter for *Read sandboxing*.

## filter/stat

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to filter for *Stat sandboxing*.

## filter/write

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to filter for *Write sandboxing*.

## filter/ioctl

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to filter for *Ioctl sandboxing*.

## filter/net/bind

[[ type
:[ *string-array*

Specifies a list of network address patterns to filter for *Bind network sandboxing*.

## filter/net/connect

[[ type
:[ *string-array*

Specifies a list of network address patterns to filter for *Connect network sandboxing*.

## stop/exec

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to stop for *Exec sandboxing*.

## stop/read

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to stop for *Read sandboxing*.

## stop/write

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to stop for *Write sandboxing*.

## stop/stat

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to stop for *Stat sandboxing*.

## stop/ioctl

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to stop for *Ioctl sandboxing*.

## stop/net/bind

[[ type
:[ *string-array*

Specifies a list of network address patterns to stop for *Bind network sandboxing*.

## stop/net/connect

[[ type
:[ *string-array*

Specifies a list of network address patterns to stop for *Connect network sandboxing*.

## kill/exec

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to kill for *Exec sandboxing*.

## kill/read

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to kill for *Read sandboxing*.

## kill/write

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to kill for *Write sandboxing*.

## kill/stat

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to kill for *Stat sandboxing*.

## kill/ioctl

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to kill for *Ioctl sandboxing*.

## kill/net/bind

[[ type
:[ *string-array*

Specifies a list of network address patterns to kill for *Bind network sandboxing*.

## kill/net/connect

[[ type
:[ *string-array*

Specifies a list of network address patterns to kill for *Connect network sandboxing*.

## exit/exec

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to exit for *Exec sandboxing*.

## exit/read

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to exit for *Read sandboxing*.

## exit/write

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to exit for *Write sandboxing*.

## exit/stat

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to exit for *Stat sandboxing*.

## exit/ioctl

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to exit for *Ioctl sandboxing*.

## exit/net/bind

[[ type
:[ *string-array*

Specifies a list of network address patterns to exit for *Bind network sandboxing*.

## exit/net/connect

[[ type
:[ *string-array*

Specifies a list of network address patterns to exit for *Connect network sandboxing*.

## append

[[ type
:[ *string-array*

Specifies a list of _glob_(3p) patterns to files that should be made
append-only for *Write sandboxing*.

If a path is append-only, Syd adds "O_APPEND" and removes "O_TRUNC" from flags
on any sandbox granted attempt to _open_(2) this path. Moreover, any attempt to
_rename_(2), _truncate_(2) and _unlink_(2) the file is prevented. This is
typically useful for history and log files.

## mask

[[ type
:[ *string-array*
|  default
:  _["/proc/cmdline", "/proc/kmsg", "/dev/kmsg"]_

Specifies a list of _glob_(3p) patterns to mask for *Read & Write sandboxing*.

If a path is masked, Syd returns a file descriptor to "/dev/null" on any
sandbox granted attempt to _open_(2) this path. Masking can effectively
be used to hide the contents of a file in a more relaxed and compatible
way than denying read/write access to it. _stat_(2) calls on a masked
file returns the original file metadata and a masked file may be
executed.

By default the list of masked paths includes "/proc/cmdline". Adhering
to the goal to be secure by default, Syd does not grant access to the
contents of this file, which may potentially contain sensitive
information such as the path to an otherwise hidden path. Experience
shows many programs expect this file to be readable so Syd prefers to
mask this file rather than denying read/write access to it, taking
advantage of the fact that an empty file is a valid "/proc/cmdline" boot
parameters file. Should the need arise to pass boot parameters inside a
Syd sandbox, this restriction may be removed with "mask-/proc/cmdline".

As of version 3.21.0, Syd adds the items "/dev/kmsg", and "/proc/kmsg"
to this list by default. This is done to provide an empty buffer to the
_dmesg_(1) tool which would otherwise fallback to using the denylisted
_syslog_(2) system call and fail with "ENOSYS", aka "Function not
implemented". Should the need arise to use these files inside a Syd
sandbox, such as when containing _syslogd_(8), these restrictions may be
removed with "mask-/dev/kmsg", and "mask-/dev/kmsg".

## cmd/exec

[[ type
:[ *command*

Makes Syd execute an external command without sandboxing. The process is
executed in a new process group with its standard input closed. The
program name and arguments must be separated with the *US* (unit
separator, octal: 037) character. To ease usage, the _syd-exec_(1)
helper utility is provided to construct a sandbox command of this type:

```
;  syd -puser -- bash -c 'test -c $(syd-exec echo hello world)'
hello world
{"cfg":"cmd/exec!echo\u001fhello\u001fworld","cmd":"bash ...}
;
```

## load

[[ type
:[ *integer*

Read configuration from the given file descriptor, the file must be open for
reading. Syd uses _pidfd_getfd_(2) to acquire the file descriptor and reads
sandbox configuration from it. This command is useful to load a set of sandbox
commands into Syd in a single step and is typically used with *reset*, e.g:

```
int fd = open("/tmp", O_RDWR | O_TMPFILE | O_CLOEXEC, 0);
if (fd == -1) errx(1, "Failed to open temporary file");

const char *syd = "sandbox:stat/on\\nallow/stat+/***\\ndeny/stat+/\\nlock:on\\n";
errx(write(fd, syd, strlen(syd)) == -1, "Failed to write config");
errx(lseek(fd, 0, SEEK_SET) == -1, "Failed to seek in file");

char load[64];
sprintf(load, "/dev/syd/load/%d", fd);
errx(stat("/dev/syd/reset", NULL) == -1, "Failed to reset syd");
errx(stat(load, NULL) == -1, "Failed to load syd profile");

errx(execvp("/bin/sh", (char *[]){"/bin/sh", "-l", NULL}) == -1, "execvp failed");
```

Due to security reasons, this command is only available via the virtual stat
call, it may not be used with the *-m* command line switch or in a configuration
file.

## trace/allow_safe_setuid

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether the Linux capability "CAP_SETUID" should be retained.
This option in combination with *SafeSetID* allows the sandbox process
to change UID. Note, Syd will change its UID with the sandbox process.

Note, because NPTL uses real-time signals to ensure all threads share
the same UID/GID, setting this option disables the SROP mitigator. See
the "Enhanced Execution Control (EEC)" section of the _syd_(7) manual
page for more information.

## trace/allow_safe_setgid

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether the Linux capability "CAP_SETGID" should be retained.
This option in combination with *SafeSetID* allows the sandbox process
to change GID. Note, Syd will change its GID with the sandbox process.

Note, because NPTL uses real-time signals to ensure all threads share
the same UID/GID, setting this option disables the SROP mitigator. See
the "Enhanced Execution Control (EEC)" section of the _syd_(7) manual
page for more information.

## setuid

[[ type
:[ *[(uid, uid)]*

Add, remove a UID transition or reset UID transitions. Only a single transition
from a source UID can be defined. Transitions to root are not allowed.

Usage:

```
setuid+0:65534      # Define a UID transition from root to nobody.
setuid+root:nobody  # Same as above but using user names.
setuid-0:65534      # Remove a previously defined UID transition.
setuid^0            # Remove all UID transitions matching source UID.
setuid^             # Remove all UID transitions.
```

## setgid

[[ type
:[ *[(gid, gid)]*

Add, remove a GID transition or reset GID transitions. Only a single transition
from a source GID can be defined. Transitions to root are not allowed.

Usage:

```
setgid+0:65534      # Define a GID transition from root to nogroup.
setgid+root:nogroup # Same as above but using group names.
setgid-0:65534      # Remove a previously defined GID transition.
setgid^0            # Remove all GID transitions matching source GID.
setgid^             # Remove all GID transitions.
```

## trace/allow_unsafe_debug

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether debugging and tracing facilities should be allowed within
the sandbox. This option is provided as a general way to allow the following:

. Disable Memory-Deny-Write-Execute protections for the sandbox process.
  This is necessary to run tools such as _valgrind_(1) which map
  executable memory in the sandbox.
. Specify whether _perf_event_open_(2) call should be allowed. This is
  necessary to run the tool _perf_(1) in the sandbox.
. Specify whether _ptrace_(2) calls should be allowed. See *syd-ls
  ptrace* for the list of process trace system calls. When set, syd does
  not drop the capability _CAP_SYS_PTRACE_ on startup. This is necessary
  to run tools like _gdb_(1) and _strace_(1) in the sandbox.
. Allow access to the kernel debugging and tracing interfaces under
  "/sys/kernel/debug/\*", and "/sys/kernel/tracing/\*". By default,
  these paths are hidden from the sandbox.
. Allow access to the unsafe files "/proc/kallsyms", "/proc/modules",
  "/proc/version" and "/sys/kernel/notes". Access to these files are
  necessary to make _perf_(1) function correctly. By default, these files
  are hidden from the sandbox.
. Allow access to the unsafe process-specific files
  "/proc/self/{auxv,environ,maps,mem,pagemap,personality,stack,syscall,wchan}".
  Access to these files are necessary for various debugging and tracing
  tools such as _gdb_(1), _strace_(1), _valgrind_(1).a By default, these
  files are hidden from the sandbox.

*Warning*, this option implies "trace/allow_unsafe_exec:1", and
"trace/allow_unsafe_memory:1", therefore disables Force Sandboxing,
SegvGuard, the exec-TOCTOU mitigator and Memory-Deny-Write-Execute
protections. This is why, *this option should only be used in trusted
environments.* Note, access granted to the debugging paths mentioned
above only applies to the unconditional unsafe path checker. These paths
remain subject to sandbox restrictions, meaning that settings like
"deny/read+/proc/\*/mem" will still prevent access to "/proc/self/mem".

## trace/allow_unsafe_dumpable

[[ type
:[ *boolean*
|  static
:  *yes*

A boolean specifying whether Syd should skip from setting its process
dumpable attribute to false. This allows core dumps for the Syd process,
and allows debugging/profiling/tracing the Syd process. You should not
set this option unless you're developing Syd.

## trace/allow_unsafe_exec

[[ type
:[ *boolean*
|  static
:  *yes*

A boolean specifying whether _ptrace_(2) should be used to secure the
exec handler. Setting this option to true effectively removes the
_ptrace_(2) dependency from the sandbox. This is necessary to trace syd
together with its children, e.g. with "strace -f". *Warning*, this
option makes _syd_(1) keep the "CAP_SYS_PTRACE" capability and disables
Force Sandboxing, SegvGuard and the exec-TOCTOU mitigator. It allows the
sandbox process to trivially break out of the sandbox by e.g. attaching
to the _syd_(1) main thread with _ptrace_(2) and getting a handle to the
_seccomp_(2) notify file descriptor. Therefore, *this option should
only be used in trusted environments.*

## trace/allow_unsafe_nopie

[[ type
:[ *boolean*

A boolean specifying whether execution of non-PIE binaries should be
allowed. This is generally not recommended but may be necessary on some
systems.

## trace/allow_unsafe_namespace

[[ type
:[ *string-array*
|  static
:  *yes*

Comma separated list of namespace types to allow creation of. Possible
values are "mount", "uts", "ipc", "user", "pid", "net", "cgroup", and
"time". If "mount" is specified, Syd allows the system calls _mount_(2),
_mount_setattr_(2), _umount_(2), and _umount2_(2) in the sandbox. This
can allow the sandbox process to circumvent path sandboxing and should
be used with extreme care.

## trace/allow_unsafe_chown

[[ type
:[ *boolean*
|  static
:  *yes*

Makes Syd keep the capability "CAP_CHOWN" and sandbox process will
inherit the capability from Syd.

## trace/allow_unsafe_open

[[ type
:[ *boolean*

A boolean specifying whether the mitigation to turn "O_PATH" file
descriptors into "O_RDONLY" file descriptors for safe emulation should
be disabled. With this option, syd continues the _open_(2) system calls
with the "O_PATH" in the sandbox process which opens a TOCTOU vector.

## trace/allow_unsafe_cpu

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether CPU emulation system calls should be allowed. By
default, as of version 3.22.1, Syd denies the _modify_ldt_(2),
_subpage_prot_(2), _switch_endian_(2), _vm86_(2), and _vm86old_(2)
system calls, which are associated with CPU emulation functionalities.
Enabling this option (trace/allow_unsafe_cpu:1) permits these calls,
thus relaxing the restriction. This option should be used with caution,
as allowing these system calls can introduce potential vulnerabilities
by enabling processes to modify CPU state or memory protections. Use
this setting only in trusted environments where the execution of these
system calls is necessary.

## trace/allow_unsafe_keyring

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether the _add_key_(2), _keyctl_(2), and _request_key_(2)
system calls should be allowed. Enabling this setting permits key
management within the sandbox, which can introduce security risks by
allowing keyring manipulations. Use only in trusted environments.

## trace/allow_unsafe_pkey

[[ type
:[ *boolean*
|  static
:  *yes*

Specifies whether the _pkey_alloc_(2), _pkey_free_(2), and
_pkey_mprotect_(2) system calls should be allowed. By default, these
calls are denied to enhance security. Setting this option to true
enables these system calls, allowing the use of memory protection keys.
This option should be used with caution and only in trusted environments
where the use of these system calls is necessary.

## trace/allow_unsafe_msgsnd

[[ type
:[ *boolean*
|  static
:  *yes*

Specifies whether the _msgsnd_(2) system call should be allowed. By
default, this call is denied to enhance security as the ability of this
system call to allocate large, contiguous blocks of memory in the kernel
heap is often used to orchestrate kernel heap spraying attacks. See the
"Mitigation Against Heap Spraying" section of the _syd_(7) manual page
for more information.

## trace/allow_unsafe_time

[[ type
:[ *boolean*
|  static
:  *yes*

A boolean specifying whether system calls which adjust the system time are
allowed. Note, this also causes Syd to keep the *CAP_SYS_TIME* capability.
Use *syd-ls time* to see the list of system calls allowed by this setting.

## trace/allow_unsafe_uring

[[ type
:[ *boolean*
|  static
:  *yes*

A boolean specifying whether system calls of the _io_uring_(7) interface are
allowed. Normally, these are denied because they may be used to bypass path
sandboxing. Use *syd-ls uring* to see the list of system calls allowed by
this setting.

## trace/allow_unsafe_caps

[[ type
:[ boolean
|  static
:  yes

A boolean specifying whether Syd should skip dropping Linux capabilities at
startup. See *syd-ls drop* for the list of capabilities Syd drops. This setting
can be used to construct _privileged containers_ and should be used with extreme
care.

Note, _syd-oci_(1) sets this option to honour the list of capabilities
specified by the container engine. You may unset it using the container
configuration file. See CONFIGURATION section in _syd-oci_(1) manual
page for more information.

## trace/allow_unsafe_env

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether unsafe environment variables should be allowed into the
environment of the sandbox process. See *syd-ls env* for the list of unsafe
environment variables.

## trace/allow_safe_kcapi

[[ type
:[ *boolean*

Specify whether access to the Linux kernel cryptography API (aka:
"KCAPI") should be allowed when network sandboxing is on. This option
has no effect when network sandboxing is off.

As most things in life, cryptography has good and evil uses: KCAPI is
convenient as it may be used to implement cryptography without depending
on user-space libraries such as OpenSSL but it may also enable malicious
code to efficiently turn itself into ransomware. Adhering to the goal to
be secure by default Syd disallows this access by default.

Note, Syd does not hook into _setsockopt_(2) and the "ALG_SET_KEY"
operation to set the encryption key is directly handled by the host
kernel therefore the encryption key is not copied into Syd's address
space.

Note again, Syd hooks into _bind_(2), _sendto_(2), _sendmsg_(2), and
_sendmmsg_(2) but *not* _read_(2), _write_(2), _recv_(2), or
_splice_(2). To reduce syscall overhead, user is recommended to use the
unhooked system calls when they can to interact with KCAPI.

## trace/allow_safe_bind

[[ type
:[ *boolean*

Specify whether the socket address arguments of successful _bind_(2) calls
should be allowed for _connect_(2), _sendto_(2), and _recvmsg_(2) system calls.
*Note*, these addresses are allowed globally and not per-process for usability
reasons. Thus, for example, a process which forks to call _bind_(2) will have
its address allowed for their parent as well.

## trace/allow_unsafe_bind

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether the Linux capability "CAP_NET_BIND_SERVICE", which allows a
process to _bind_(2) to ports lower than 1024, should be retained.

## trace/allow_unsafe_socket

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether unsafe socket families should be allowed. When set, Syd does
not drop the capability _CAP_NET_RAW_ on startup. This allows:

. use of RAW and PACKET sockets;
. bind to any address for transparent proxying.

## trace/allow_unsupp_socket

[[ type
:[ *boolean*

Specify whether unsupported socket families such as netlink sockets
should be allowed access when network sandboxing is on. By default Syd
allows sandboxed access to unix, ipv4 and ipv6 sockets. This option has
no effect when network sandboxing is off.

As of version 3.16.6 Syd allows access to alg sockets with the
"trace/allow_safe_kcapi" option rather than with this option. Alg
sockets are used to interact with the Linux kernel cryptography API.

Note, on architectures with a multiplexed _socketcall_(2) system call,
enabling this option is insecure because it is vulnerable to TOCTOU.
You may use _syd-sys_(1) utility to check if this system call is
available on your architecture using e.g: "syd-sys socketcall && echo
vulnerable".

## trace/allow_unsafe_prctl

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether _prctl_(2) restrictions should be lifted. See *syd-ls prctl* for
the list of allowed prctl requests.

## trace/allow_unsafe_prlimit

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether _prlimit_(2) restrictions should be lifted.

## trace/allow_unsafe_syslog

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether the Linux capability "CAP_SYSLOG" should be retained. This
allows the process to perform privileged _syslog_(2) operations. This is
useful when sandboxing a service such as syslogd.

## trace/allow_unsafe_memfd

[[ type
:[ *boolean*

A boolean specifying whether secret memory file descriptors and
executable memory file descriptors should be enabled. By default Syd
strips the "MFD_EXEC" and adds the "MFD_NOEXEC_SEAL" flag to
_memfd_create_(2) flags argument. This ensures the memory file
descriptor can never be made executable. The "MFD_NOEXEC_SEAL" flag
requires Linux-6.3 or newer therefore on older kernels this option must
be enabled to make memory file descriptors work. However, the user
should be aware that allowing encrypted memory file descriptors does
allow an attacker to bypass Exec, Force and TPE sandboxing and execute
denylisted code.

## trace/allow_unsafe_memory

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether the Memory-Deny-Write-Execute (MDWE) protections should be
bypassed. See *Memory-Deny-Write-Execute Protections* section of the _syd_(7)
manual page for more information.

## trace/deny_dotdot

[[ type
:[ *boolean*

Specify whether ".." components should be denied during path resolution for
_chdir_(2) and _open_(2) family system calls. This is useful in mitigating path
traversal attacks. See "Path Resolution Restriction For Chdir and Open Calls" of
the _syd_(7) manual page for more information.

## trace/deny_elf32

[[ type
:[ *boolean*

Deny the execution of 32-bit ELF binaries.

## trace/deny_elf_dynamic

[[ type
:[ *boolean*

Deny the execution of dynamically linked ELF binaries.

## trace/deny_elf_static

[[ type
:[ *boolean*

Deny the execution of statically linked ELF binaries.

## trace/deny_script

[[ type
:[ *boolean*

Deny the execution of scripts (files with #!<interpreter> on first line).

Note, the _execve_(2) TOCTOU mitigations do not cover this option which
means the functionality is vulnerable to TOCTOU. This allows an attacker
to execute a script whose path is denylisted. This TOCTOU is limited to
scripts and requires the interpreter binary to be allowlisted for exec.
Hence this vulnerability does not allow an attacker to execute
denylisted binaries. This is why the user is recommended to deny the
respective interpreter binaries for execution instead for a safe and
secure approach.

## trace/deny_tsc

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether reading the timestamp counter should be denied. Without an
accurate timer, many timing attacks are going to be harder to perform.

- This works on *x86 only*.
- This breaks time related calls in the *vDSO*, which can be trivially worked \
around by writing a *LD_PRELOAD* library to call the respective system calls \
directly. See *libsydtime*, https://lib.rs/libsydtime, for a reference \
implementation.
- This has a negative performance impact on programs that rely on \
_gettimeofday_(2) being a *vDSO* call.

## trace/force_cloexec

[[ type
:[ *boolean*

Specify whether the "O_CLOEXEC" flag should be enforced for all
_open_(2), _openat_(2), and _openat2_(2) system calls made by the
sandbox process. When this feature is enabled, Syd ensures that every
file descriptor opened by the sandbox process is automatically set with
the "O_CLOEXEC" flag, which prevents these file descriptors from being
inherited by newly executed programs. This measure enhances security by
closing file descriptors during _exec_(3) calls, thereby mitigating the
risk of file descriptor leakage which could lead to unauthorized access
to sensitive files or resources. The feature can be toggled at runtime
using Syd's virtual stat API, providing flexible control over the
confinement level of sandboxed processes.

## trace/lock_personality

[[ type
:[ *boolean*
|  static
:  *yes*

If set, locks down the _personality_(2) system call so that the kernel execution
domain may not be changed from the default. This may be useful to improve
security, because odd personality emulations may be poorly tested and source of
vulnerabilities.

## trace/exit_wait_all

[[ type
:[ *boolean*
|  static
:  *yes*

Specify whether Syd should wait for all processes to exit before exiting. By
default, Syd exits with the eldest process and any leftover processes in the
background are automatically killed.

## trace/force_umask

[[ type
:[ *octal*
|  static
:  *yes*

Specify an umask mode to force. To unset a previously configured force
umask use -1 as the value. As of version 3.15.6, _chmod_(2) family
system calls also honour force umask for added hardening. As of version
3.22.1, this setting does not apply to directory creation for _mkdir_(2)
and _mkdirat_(2) system calls.

## trace/memory_access

[[ type
:[ *integer*
|  default
:  *0*
|  static
:  *yes*

Set mode on cross memory attach and */proc/pid/mem* usage. Cross memory attach
requires a Linux kernel with the *CONFIG_CROSS_MEMORY_ATTACH* option enabled.
Supported modes are:

- *0*: Use cross memory attach if available, use */proc/pid/mem* otherwise.
- *1*: Use */proc/pid/mem* unconditionally.

```
;  strace -q -eprocess_vm_readv -fc -- syd -m trace/memory_access:1 true
[pid 1100565] ????( <detached ...>
;  strace -q -eprocess_vm_readv -fc -- syd -m trace/memory_access:0 true
% time     seconds  usecs/call     calls    errors syscall
------ ----------- ----------- --------- --------- -----------------
100,00    0,000155           6        24           process_vm_readv
------ ----------- ----------- --------- --------- -----------------
100,00    0,000155           6        24           total
;
```

## cache/path/icap

[[ type
:[ *usize*
|  default
:  *1024*
|  static
:  *yes*

Estimated item capacity of the security policy cache for paths.

## cache/path/wcap

[[ type
:[ *u64*
|  default
:  *36684*
|  static
:  *yes*

Weight capacity of the security policy cache for paths.

## cache/addr/icap

[[ type
:[ *usize*
|  default
:  *4096*
|  static
:  *yes*

Estimated item capacity of the security policy cache for IP addresses.

## cache/addr/wcap

[[ type
:[ *u64*
|  default
:  *65536*
|  static
:  *yes*

Weight capacity of the security policy cache for IP addresses.

# SHORTCUTS

*read*, *stat*, *write*, and *exec* may be passed to *allow*, *deny*, and
*filter* either as a single unit or as a comma-delimited list, e.g:

```
allow/read,write,stat,exec+/***
allow/read,write,stat-/***
deny/read,write,stat+/***
deny/read,write-/***
filter/read,write,stat+/dev/mem
filter/read,write-/dev/mem
```

Since 3.18.14, sandboxing modes may be specified as a single unit or as a
comma-delimited list, e.g:

```
sandbox/read,write,stat,exec:on
sandbox/net,lock:off
```

Since 3.19.0, namespace types may be specified as a single unit or as a
comma-delimited list, e.g.:

```
unshare/user,pid,mount:on
unshare/net,cgroup:off
```

# PATTERN MATCHING

Syd uses shell-style pattern matching for allowlists and filters. The
matching code uses the *globset* crate. Check their documentation for more
information about patterns. Note, patterns are _case sensitive_, the *empty
alternates* option is set when building patterns, i.e. *foo{,.txt}* in a pattern
will match both *foo* and *foo.txt* and Syd applies the *triple star*
extension to patterns, i.e. */dev/\*\*\** matches both */dev* and any file
recursively under */dev*. Note also, Syd gets patterns from multiple sources:
a configuration file, a profile, the *-m* command line switch, or a *stat* call
with */dev/syd* prefix. There is no precedence between different sources. All
patterns in a list are compiled together in a *GlobSet* and pattern matching
during access control happens in a single step where _the last matching pattern
decides the outcome._

# ADDRESS MATCHING

Syd has a simple address scheme to match network addresses. The addresses can
either be a *glob* pattern to match *UNIX* and *abstract UNIX* socket addresses,
or *IP CIDR* followed by a port range to match *IPv4* and *IPv6* addresses. Port
range can either be a single port or a range in format *port1-port2*. The
address and the port range must be split by the character *!*.  The precedence
logic is same as *Pattern Matching* where _the last matching pattern decides the
outcome._

In addition there are some *aliases*, you may use instead of specifying an address:

- *any*: Expanded to *any4* + *any6*.
- *any4*: Expanded to *0.0.0.0/0* which matches the whole Ipv4 address space.
- *any6*: Expanded to *::/0* which matches the whole Ipv6 address space.
- *loopback*: Expanded to *loopback4* + *loopback6*.
- *loopback4*: Expanded to *127.0.0.0/8*
- *loopback6*: Expanded to *::1/128*
- *linklocal*: Expanded to *linklocal4* + *linklocal6*.
- *linklocal4*: Expanded to *169.254.0.0/16*
- *linklocal6*: Expanded to *fe80::/10*
- *local*: Expanded to *local4* + *local6*.
- *local4*: Expanded to four addresses as defined in *RFC1918*:
	- *127.0.0.0/8*
	- *10.0.0.0/8*
	- *172.16.0.0/12*
	- *192.168.0.0/16*
- *local6*: Expanded to four addresses:
	- *::1/128*
	- *fe80::/7*
	- *fc00::/7*
	- *fec0::/7*

# SECURITY

The interface is only available if the sandbox lock is not set for the calling
process. Similarly, command-line option parsing and configuration file parsing
stops once a *lock:on* clause is executed.

# RETURN VALUE

For _stat_(2) calls, on success, zero is returned. On error, -1 is returned, and
errno is set to indicate the error.

For _open_(2) calls, on success the new file descriptor (a nonnegative integer)
is returned. On error, -1 is returned and errno is set to indicate the error.

# ERRORS

|[ *EBUSY*
:< Attempted to edit a setting at runtime that must be configured at startup

|[ *EEXIST*
:< Attempted to add an UID/GID transition for SafeSetID but a transition with the same source UID/GID exists.

|[ *EINVAL*
:< Sandbox command is syntactically incorrect.

|[ *ENOENT*
:< Result of the given sandbox query is false (e.g. "test -c /dev/syd/sandbox/stat?").

|[ *ENOENT*
:< Sandbox lock is on, no commands are allowed.

|[ *ENOSYS*
:< elf feature is disabled and command is one of "trace/deny_elf32", "trace/deny\_dynamic", or "trace/deny\_static".

# SEE ALSO

_syd_(1), _syd_(5), _syd_(7), _syd-ls_(1) _open_(2), _stat_(2),
_perf_(1), _pledge_(2), _ptrace_(2), _seccomp_(2), _strace_(1),
_glob_(3p), _io_uring_(7), _gdb_(1), _valgrind_(1)

- *syd* homepage: https://sydbox.exherbolinux.org
- *libsyd* homepage: https://libsyd.exherbolinux.org
- *gosyd* homepage: https://gosyd.exherbolinux.org
- *plsyd* homepage:  https://plsyd.exherbolinux.org
- *pysyd* homepage: https://pysyd.exherbolinux.org
- *rbsyd* homepage: https://rbsyd.exherbolinux.org
- *syd.el* homepage: https://sydel.exherbolinux.org
- *libsydtime* homepage: https://lib.rs/libsydtime
- *LandLock* homepage: https://landlock.io/
- *vDSO* wiki: https://en.wikipedia.org/wiki/VDSO
- *parse-size* documentation: https://docs.rs/parse-size/1.0.0/parse_size/
- *globset* documentation: https://docs.rs/globset
- *GlobSet* documentation: https://docs.rs/globset/latest/globset/struct.GlobSet.html
- GlobBuilder *empty_alternates* method: https://docs.rs/globset/latest/globset/struct.GlobBuilder.html#method.empty_alternates
- *ipnetwork* documentation: https://docs.rs/ipnetwork

# AUTHORS

Maintained by Ali Polatel. Up-to-date sources can be found at
https://gitlab.exherbo.org/sydbox/sydbox.git and bugs/patches can be
submitted to https://gitlab.exherbo.org/groups/sydbox/-/issues. Discuss
in #sydbox on Libera Chat.
